# yggdrasil.io.response
"""HTTP response model with Arrow, Polars, pandas, and ASGI serialisation."""
from __future__ import annotations

import datetime as dt
import warnings
from dataclasses import MISSING, dataclass, replace, field
from typing import TYPE_CHECKING, Any, Callable, Iterable, Iterator, Literal, Mapping, MutableMapping, Optional

import pyarrow as pa

import yggdrasil.pickle.json as json_module
from yggdrasil.data import any_to_datetime
from yggdrasil.dataclasses.dataclass import get_from_dict
from .buffer import BytesIO
from .enums import Codec, MediaType, MimeType
from .headers import DEFAULT_HOSTNAME, PromotedHeaders, normalize_headers
from .request import PreparedRequest, REQUEST_ARROW_SCHEMA, _build_metadata, _field

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    from fastapi import Response as FastAPIResponse
    from pyspark.sql import DataFrame as SparkDataFrame, Row as SparkRow
    from starlette.responses import Response as StarletteResponse


__all__ = [
    "Response",
    "BASE_ARROW_SCHEMA",
    "RESPONSE_ARROW_SCHEMA",
]


# ---------------------------------------------------------------------------
# Private header / body helpers
# ---------------------------------------------------------------------------

def _get_header(headers: Mapping[str, str] | None, name: str) -> str | None:
    if not headers:
        return None

    value = headers.get(name)
    if value is not None:
        return str(value)

    target = name.lower()
    for key, value in headers.items():
        if isinstance(key, str):
            if key == name or key.lower() == target:
                return str(value)
        elif str(key).lower() == target:
            return str(value)

    return None


def _pop_header(headers: MutableMapping[str, str], name: str) -> str | None:
    value = headers.pop(name, None)
    if value is not None:
        return str(value)

    target = name.lower()
    for key in list(headers.keys()):
        if isinstance(key, str):
            if key == name or key.lower() == target:
                return str(headers.pop(key))
        elif str(key).lower() == target:
            return str(headers.pop(key))

    return None


def _get_charset(headers: Mapping[str, str]) -> str:
    content_type = _get_header(headers, "Content-Type")
    if not content_type:
        return "utf-8"

    for part in str(content_type).split(";")[1:]:
        part = part.strip()
        if part.lower().startswith("charset="):
            charset = part.split("=", 1)[1].strip().strip('"')
            return charset or "utf-8"

    return "utf-8"


def _parse_content_type(headers: Mapping[str, str] | None) -> str | None:
    value = _get_header(headers, "Content-Type")
    if not value:
        return None
    bare = str(value).split(";", 1)[0].strip().lower()
    return bare or None


def _parse_content_encoding(headers: Mapping[str, str] | None) -> str | None:
    value = _get_header(headers, "Content-Encoding")
    if not value:
        return None
    parts = [p.strip().lower() for p in str(value).split(",") if p.strip()]
    return ",".join(parts) or None


def _parse_content_length(headers: Mapping[str, str] | None) -> int | None:
    value = _get_header(headers, "Content-Length")
    if value in (None, ""):
        return None
    try:
        return int(str(value).strip())
    except Exception:
        return None


def _is_probably_placeholder_content_type(value: str | None) -> bool:
    if not value:
        return True
    return value.strip().lower() in {
        "",
        "application/octet-stream",
        "binary/octet-stream",
        "unknown/unknown",
    }


def _sniff_media_from_body(
    body: BytesIO,
    *,
    content_type: str | None,
    content_encoding: str | None,
) -> MediaType:
    content_codec = Codec.parse(content_encoding, default=None)

    if content_type and not _is_probably_placeholder_content_type(content_type):
        try:
            return MediaType.parse(content_type, codec=content_codec)
        except Exception:
            pass

    try:
        sniffed = body.media_type
    except Exception:
        sniffed = MediaType(MimeType.OCTET_STREAM, content_codec)

    return MediaType(sniffed.mime_type, codec=content_codec) if content_encoding else sniffed


def _ensure_media_headers(
    headers: MutableMapping[str, str],
    body: BytesIO,
) -> MediaType:
    declared_type = _parse_content_type(headers)
    declared_encoding = _parse_content_encoding(headers)

    media = _sniff_media_from_body(
        body,
        content_type=declared_type,
        content_encoding=declared_encoding,
    )

    if _is_probably_placeholder_content_type(declared_type):
        headers["Content-Type"] = media.mime_type.value

    if not declared_encoding and media.codec is not None:
        headers["Content-Encoding"] = media.codec.name

    if _parse_content_length(headers) is None:
        headers["Content-Length"] = str(body.size)

    return media


def _string_dict(arg: Optional[Mapping[Any, Any]]) -> dict[str, str]:
    if not arg:
        return {}
    return {str(k): str(v) for k, v in arg.items()}


def _map_to_str_dict(value: Any) -> dict[str, str]:
    if not value:
        return {}
    if isinstance(value, Mapping):
        return {str(k): str(v) for k, v in value.items() if k is not None and v is not None}
    try:
        return {str(k): str(v) for k, v in value if k is not None and v is not None}
    except Exception:
        return {}


# ---------------------------------------------------------------------------
# Parsing helpers
# ---------------------------------------------------------------------------

def _parse_headers(obj: Mapping[str, Any], *, prefix: str) -> MutableMapping[str, str]:
    headers = get_from_dict(obj, keys=("headers", "header", "hdrs", "response_headers"), prefix=prefix)
    if headers is MISSING:
        headers = get_from_dict(obj, keys=("headers", "header", "hdrs", "response_headers"), prefix="")

    if isinstance(headers, Mapping):
        parsed = _string_dict(headers)
        if parsed:
            return parsed

    dumped_headers = {
        "Host": get_from_dict(obj, keys=("host",), prefix=prefix),
        "User-Agent": get_from_dict(obj, keys=("user_agent",), prefix=prefix),
        "Accept": get_from_dict(obj, keys=("accept",), prefix=prefix),
        "Accept-Encoding": get_from_dict(obj, keys=("accept_encoding",), prefix=prefix),
        "Accept-Language": get_from_dict(obj, keys=("accept_language",), prefix=prefix),
        "Content-Type": get_from_dict(obj, keys=("content_type",), prefix=prefix),
        "Content-Length": get_from_dict(obj, keys=("content_length",), prefix=prefix),
        "Content-Encoding": get_from_dict(obj, keys=("content_encoding",), prefix=prefix),
        "Transfer-Encoding": get_from_dict(obj, keys=("transfer_encoding",), prefix=prefix),
        "Location": get_from_dict(obj, keys=("location",), prefix=prefix),
        "ETag": get_from_dict(obj, keys=("etag",), prefix=prefix),
        "Last-Modified": get_from_dict(obj, keys=("last_modified",), prefix=prefix),
    }

    return {
        header_name: str(value)
        for header_name, value in dumped_headers.items()
        if value is not MISSING and value not in (None, "")
    }


def _parse_tags(obj: Mapping[str, Any], *, prefix: str) -> dict[str, str]:
    tags = get_from_dict(obj, keys=("tags", "response_tags"), prefix=prefix)
    if tags is MISSING:
        tags = get_from_dict(obj, keys=("tags", "response_tags"), prefix="")
    return _string_dict(tags if isinstance(tags, Mapping) else None)


def _parse_buffer(obj: Mapping[str, Any], *, prefix: str) -> BytesIO:
    body = get_from_dict(obj, keys=("buffer", "body", "content", "data", "response_body"), prefix=prefix)
    if body is MISSING:
        body = get_from_dict(obj, keys=("buffer", "body", "content", "data", "response_body"), prefix="")
    if body is MISSING or body is None:
        return BytesIO()
    return BytesIO.parse(obj=body)


def _parse_status_code(obj: Mapping[str, Any], *, prefix: str) -> int:
    status = get_from_dict(obj, keys=("status_code", "status", "code"), prefix=prefix)
    if status is MISSING:
        status = get_from_dict(obj, keys=("status_code", "status", "code"), prefix="")
    if status is MISSING or status in (None, ""):
        raise ValueError("Response.parse_mapping: missing status_code/status/code")
    return int(status) if isinstance(status, int) else int(float(str(status).strip()))


def _parse_received_at(obj: Mapping[str, Any], *, prefix: str) -> int:
    keys = (
        "received_at_timestamp",
        "received_at",
        "response_received_at",
    )
    value = get_from_dict(obj, keys=keys, prefix=prefix)
    if value is MISSING:
        value = get_from_dict(obj, keys=keys, prefix="")

    if value is MISSING:
        return dt.datetime.fromtimestamp(0, tz=dt.timezone.utc)
    return any_to_datetime(value)


def _arrow_ts_col_to_us(col: pa.ChunkedArray | pa.Array, i: int) -> int:
    scalar = col[i]
    if scalar is None or not scalar.is_valid:
        return 0
    value = scalar.as_py()
    return any_to_datetime(value)


def _first_present(cols: Mapping[str, Any], i: int, *names: str) -> Any:
    for name in names:
        if name in cols:
            return cols[name][i].as_py()
    return None


# ---------------------------------------------------------------------------
# Hop-by-hop header names
# ---------------------------------------------------------------------------

_HOP_BY_HOP: frozenset[str] = frozenset({
    "connection",
    "keep-alive",
    "proxy-authenticate",
    "proxy-authorization",
    "te",
    "trailers",
    "transfer-encoding",
    "upgrade",
})


# ---------------------------------------------------------------------------
# Arrow schemas
# ---------------------------------------------------------------------------

_RESPONSE_BASE_SCHEMA_JSON_TAGS: dict[str, str] = {
    "domain": "http",
    "entity": "response",
    "layer": "bronze",
    "namespace": "yggdrasil.io.response",
}

_RESPONSE_COMBINED_SCHEMA_JSON_TAGS: dict[str, str] = {
    "domain": "http",
    "entity": "response",
    "layer": "bronze",
    "namespace": "yggdrasil.io.response"
}

BASE_ARROW_SCHEMA = pa.schema(
    [
        _field(
            "response_status_code",
            pa.int32(),
            nullable=False,
            comment="HTTP status code returned by the server",
            json_tags={
                "entity": "response",
                "group": "status",
                "semantic_type": "http_status_code",
                "dimension": "low_cardinality",
            },
        ),
        _field(
            "response_host",
            pa.string(),
            nullable=True,
            comment="HTTP Host header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
                "semantic_type": "host_header",
            },
        ),
        _field(
            "response_user_agent",
            pa.string(),
            nullable=True,
            comment="HTTP User-Agent header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_accept",
            pa.string(),
            nullable=True,
            comment="HTTP Accept header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_accept_encoding",
            pa.string(),
            nullable=True,
            comment="HTTP Accept-Encoding header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_accept_language",
            pa.string(),
            nullable=True,
            comment="HTTP Accept-Language header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_content_type",
            pa.string(),
            nullable=True,
            comment="HTTP Content-Type header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_content_length",
            pa.int64(),
            nullable=False,
            comment="HTTP Content-Length header parsed as integer",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_content_encoding",
            pa.string(),
            nullable=True,
            comment="HTTP Content-Encoding header",
            json_tags={
                "entity": "response",
                "group": "headers_promoted",
            },
        ),
        _field(
            "response_transfer_encoding",
            pa.string(),
            nullable=True,
            comment="HTTP Transfer-Encoding header",
            json_tags={
                "entity": "response",
            },
        ),
        _field(
            "response_headers",
            pa.map_(pa.string(), pa.string()),
            nullable=False,
            comment="Response headers excluding promoted common headers",
            json_tags={
                "entity": "response",
                "group": "headers",
            },
            keys_sorted="false",
        ),
        _field(
            "response_tags",
            pa.map_(pa.string(), pa.string()),
            nullable=False,
            comment="Arbitrary string tags attached to this response",
            json_tags={
                "entity": "response",
                "group": "tags",
            },
        ),
        _field(
            "response_body",
            pa.binary(),
            nullable=True,
            comment="Raw binary payload of the response",
            json_tags={
                "entity": "response",
                "group": "payload",
            },
        ),
        _field(
            "response_body_hash",
            pa.int64(),
            nullable=True,
            comment="Signed Int64 XXH3 digest of response_body",
            json_tags={
                "entity": "response",
                "group": "payload",
                "algorithm": "xxh3_64",
            },
            algorithm="xxh3_64",
        ),
        _field(
            "response_received_at",
            pa.timestamp("us", "UTC"),
            nullable=False,
            comment="UTC timestamp when the response was captured",
            json_tags={
                "entity": "response",
                "group": "timing",
                "timezone": "UTC",
            },
            unit="us",
            tz="UTC",
        ),
    ],
    metadata=_build_metadata(
        comment="Response record (single row), designed for deterministic logging and replay.",
        json_tags=_RESPONSE_BASE_SCHEMA_JSON_TAGS,
    ),
)

RESPONSE_ARROW_SCHEMA = pa.schema(
    list(REQUEST_ARROW_SCHEMA) + list(BASE_ARROW_SCHEMA),
    metadata=_build_metadata(
        comment="Prepared request and response flattened into a single row schema for Delta-table caching.",
        json_tags=_RESPONSE_COMBINED_SCHEMA_JSON_TAGS,
    ),
)

_RESPONSE_FIELD_NAMES: frozenset[str] = frozenset(RESPONSE_ARROW_SCHEMA.names)
_PROMOTED_RESPONSE_HEADER_FIELDS: tuple[tuple[str, str], ...] = (
    ("Host", "response_host"),
    ("User-Agent", "response_user_agent"),
    ("Accept", "response_accept"),
    ("Accept-Encoding", "response_accept_encoding"),
    ("Accept-Language", "response_accept_language"),
    ("Content-Type", "response_content_type"),
    ("Content-Length", "response_content_length"),
    ("Content-Encoding", "response_content_encoding"),
    ("Transfer-Encoding", "response_transfer_encoding"),
)


# ---------------------------------------------------------------------------
# Response dataclass
# ---------------------------------------------------------------------------

@dataclass(slots=True)
class Response:
    request: PreparedRequest
    status_code: int
    headers: MutableMapping[str, str] = field(compare=False, hash=False, repr=False)
    tags: MutableMapping[str, str] = field(compare=False, hash=False, repr=False)
    buffer: BytesIO = field(compare=False, hash=False, repr=False)
    received_at: dt.datetime = field(compare=False, hash=False, repr=False)

    def __post_init__(self) -> None:
        self.status_code = int(self.status_code)
        self.headers = _string_dict(self.headers)
        self.tags = _string_dict(self.tags)
        self.received_at = any_to_datetime(self.received_at)

        if not isinstance(self.buffer, BytesIO):
            self.buffer = BytesIO(self.buffer, copy=False)

        _ensure_media_headers(self.headers, self.buffer)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<r={self.request} b={self.body!r}>"

    def __str__(self) -> str:
        return self.__repr__()

    # ------------------------------------------------------------------
    # Construction
    # ------------------------------------------------------------------

    @classmethod
    def parse(cls, obj: Any, *, normalize: bool = True) -> "Response":
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, str):
            return cls.parse_str(obj, normalize=normalize)
        if isinstance(obj, Mapping):
            return cls.parse_mapping(obj, normalize=normalize)
        return cls.parse_str(str(obj), normalize=normalize)

    @classmethod
    def parse_str(cls, raw: str, *, normalize: bool = True) -> "Response":
        s = raw.strip()
        if not s:
            raise ValueError("Response.parse_str: empty string")
        try:
            d = json_module.loads(s)
        except Exception as exc:
            raise ValueError("Response.parse_str: expected JSON object string") from exc
        if not isinstance(d, Mapping):
            raise ValueError("Response.parse_str: JSON must decode to a mapping")
        return cls.parse_mapping(d, normalize=normalize)

    @classmethod
    def parse_mapping(
        cls,
        obj: Mapping[str, Any],
        *,
        normalize: bool = True,
        prefix: str = "response_",
    ) -> "Response":
        if not obj:
            raise ValueError("Response.parse_mapping: empty mapping")

        req_obj = get_from_dict(obj, keys=("request",), prefix="")
        request = PreparedRequest.parse(
            obj if req_obj is MISSING or req_obj in (None, "") else req_obj,
            normalize=normalize,
        )

        status_code = _parse_status_code(obj, prefix=prefix)
        headers = _parse_headers(obj, prefix=prefix)
        buffer = _parse_buffer(obj, prefix=prefix)
        received_at = _parse_received_at(obj, prefix=prefix)
        tags = _parse_tags(obj, prefix=prefix)

        if normalize:
            headers = normalize_headers(headers, body=buffer, is_request=False)

        _ensure_media_headers(headers, buffer)

        if cls is Response:
            if request.url.is_http:
                from .http_ import HTTPResponse

                return HTTPResponse(
                    request=request,
                    status_code=status_code,
                    headers=headers,
                    tags=tags,
                    buffer=buffer,
                    received_at=received_at,
                )

        return cls(
            request=request,
            status_code=status_code,
            headers=headers,
            tags=tags,
            buffer=buffer,
            received_at=received_at,
        )

    # ------------------------------------------------------------------
    # Mutation helpers
    # ------------------------------------------------------------------

    def update_headers(
        self,
        headers: MutableMapping[str, str],
        normalize: bool = True,
    ) -> "Response":
        if not headers:
            return self

        if not self.headers:
            self.headers = _string_dict(headers)
        else:
            for k, v in headers.items():
                self.headers[str(k)] = str(v)

        if normalize:
            _ensure_media_headers(self.headers, self.buffer)

        return self

    def update_tags(
        self,
        tags: MutableMapping[str, str],
    ) -> "Response":
        if not tags:
            return self

        if not self.tags:
            self.tags = _string_dict(tags)
        else:
            self.tags.update(_string_dict(tags))

        return self

    # ------------------------------------------------------------------
    # Media type
    # ------------------------------------------------------------------

    @property
    def media_type(self) -> MediaType:
        if self.headers is None:
            self.headers = {}
        return _ensure_media_headers(self.headers, self.buffer)

    @media_type.setter
    def media_type(self, value: MediaType) -> None:
        self.set_media_type(value, safe=True)

    def set_media_type(
        self,
        value: MediaType,
        *,
        safe: bool = True,
    ) -> "Response":
        if self.headers is None:
            self.headers = {}

        self.request.accept_media_type = value
        self.buffer.set_media_type(value, safe=safe)
        self.headers["Content-Type"] = value.mime_type.value

        if value.codec is not None:
            self.headers["Content-Encoding"] = value.codec.name
        else:
            self.headers.pop("Content-Encoding", None)

        self.headers["Content-Length"] = str(self.buffer.size)
        return self

    # ------------------------------------------------------------------
    # Body accessors
    # ------------------------------------------------------------------

    @property
    def body(self) -> BytesIO:
        return self.buffer

    @property
    def codec(self) -> Optional[Codec]:
        return self.media_type.codec

    @property
    def content(self) -> bytes:
        codec = self.codec
        if codec is not None:
            with self.buffer.decompress(codec=codec, copy=True) as b:
                return b.to_bytes()
        return self.buffer.to_bytes()

    @property
    def text(self) -> str:
        return self.content.decode(_get_charset(self.headers), errors="replace")

    def json(
        self,
        orient: Optional[Literal["records", "split", "index", "columns", "values"]] = None,
        *,
        media_type: Optional[MediaType] = None,
    ) -> Any:
        return self.buffer.json_load(
            orient=orient,
            media_type=media_type or self.media_type,
        )

    # ------------------------------------------------------------------
    # Status helpers
    # ------------------------------------------------------------------

    @property
    def ok(self) -> bool:
        return 200 <= self.status_code < 400

    def raise_for_status(self) -> None:
        if not self.ok:
            raise self.error()

    def warn_for_status(self) -> None:
        err = self.error()
        if err is not None:
            warnings.warn(str(err), category=RuntimeWarning, stacklevel=2)

    def error(self) -> Optional[Exception]:
        if not self.ok:
            from .errors import make_for_status
            return make_for_status(self)
        return None

    # ------------------------------------------------------------------
    # Timestamps
    # ------------------------------------------------------------------

    @property
    def received_at_timestamp(self) -> int:
        return int(self.received_at.timestamp() * 1000000)

    # ------------------------------------------------------------------
    # Matching / projection
    # ------------------------------------------------------------------

    @property
    def arrow_values(self) -> dict[str, Any]:
        promoted = PromotedHeaders.extract(self.headers or {}, host=DEFAULT_HOSTNAME)

        if self.buffer is not None:
            body_bytes = self.buffer.to_bytes()
            body_hash = self.buffer.xxh3_int64()
        else:
            body_bytes = None
            body_hash = None

        return {
            **self.request.arrow_values,
            "response_status_code": self.status_code,
            "response_host": promoted.host or DEFAULT_HOSTNAME,
            "response_user_agent": promoted.user_agent,
            "response_accept": promoted.accept,
            "response_accept_encoding": promoted.accept_encoding,
            "response_accept_language": promoted.accept_language,
            "response_content_type": promoted.content_type,
            "response_content_length": promoted.content_length or 0,
            "response_content_encoding": promoted.content_encoding,
            "response_transfer_encoding": promoted.transfer_encoding,
            "response_headers": promoted.remaining,
            "response_tags": _string_dict(self.tags),
            "response_body": body_bytes,
            "response_body_hash": body_hash,
            "response_received_at": self.received_at,
        }

    def match_value(self, key: str) -> Any:
        values = self.arrow_values
        if key in values:
            return values[key]
        if hasattr(self, key):
            return getattr(self, key)
        if key.startswith("request_"):
            return self.request.match_value(key)
        raise ValueError(
            f"Unsupported response match key: {key!r}. "
            f"Must be within: {RESPONSE_ARROW_SCHEMA.names!r}"
        )

    def match_values(
        self,
        keys: Iterable[str],
    ) -> dict[str, Any]:
        return {str(key): self.match_value(str(key)) for key in keys}

    def match_tuple(
        self,
        keys: Iterable[str],
    ) -> tuple[Any, ...]:
        key_list = [str(key) for key in keys]
        values = self.match_values(key_list)
        return tuple(values[key] for key in key_list)

    # ------------------------------------------------------------------
    # Anonymisation
    # ------------------------------------------------------------------

    def anonymize(self, mode: Literal["remove", "redact"] = "remove") -> "Response":
        if not mode:
            return self

        return replace(
            self,
            request=self.request.anonymize(mode=mode),
            headers=normalize_headers(
                self.headers,
                is_request=False,
                mode=mode,
                body=self.body,
                anonymize=True,
            ),
        )

    # ------------------------------------------------------------------
    # Serialisation — Arrow
    # ------------------------------------------------------------------

    def to_arrow_batches(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> Iterator[pa.RecordBatch]:
        if parse:
            mio = self.buffer.media_io(media=self.media_type)
            yield from mio.read_arrow_batches(lazy=lazy, **media_options)
            return

        values = self.arrow_values
        arrays = [
            pa.array([values[f.name]], type=f.type)
            for f in RESPONSE_ARROW_SCHEMA
        ]
        yield pa.RecordBatch.from_arrays(arrays, schema=RESPONSE_ARROW_SCHEMA)

    def to_arrow_batch(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> pa.RecordBatch:
        batches = list(self.to_arrow_batches(parse=parse, lazy=lazy, **media_options))
        return pa.concat_batches(batches)

    def to_arrow_table(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> pa.Table:
        batches = list(self.to_arrow_batches(parse=parse, lazy=lazy, **media_options))
        return pa.Table.from_batches(batches)

    # ------------------------------------------------------------------
    # Serialisation — Polars / pandas / Spark
    # ------------------------------------------------------------------

    def read_polars_frames(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ):
        if parse:
            mio = self.buffer.media_io(media=self.media_type)

            for df in mio.read_polars_frames(lazy=lazy, **media_options):
                yield df
        else:
            from yggdrasil.polars.lib import polars as _pl

            yield _pl.from_arrow(self.to_arrow_batch(parse=False))

    def to_polars(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> "pl.DataFrame | pl.LazyFrame":
        from yggdrasil.polars.lib import polars as _pl

        if parse:
            mio = self.buffer.media_io(media=self.media_type)
            return mio.read_polars_frame(lazy=lazy, **media_options)

        return _pl.from_arrow(self.to_arrow_batch(parse=False))

    def to_pandas(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> "pd.DataFrame":
        return self.to_arrow_table(
            parse=parse,
            lazy=lazy,
            **media_options,
        ).to_pandas()

    def to_spark(
        self,
        parse: bool = True,
        *,
        lazy: bool = False,
        **media_options: Any,
    ) -> "SparkDataFrame":
        from yggdrasil.spark.cast import arrow_table_to_spark_dataframe

        return arrow_table_to_spark_dataframe(
            self.to_arrow_table(parse=parse, lazy=lazy, **media_options)
        )

    # ------------------------------------------------------------------
    # Arrow / Spark deserialization
    # ------------------------------------------------------------------

    @classmethod
    def from_spark_frame(
        cls,
        df: "SparkDataFrame",
    ) -> Iterator["Response"]:
        for row in df.toLocalIterator():
            yield cls.from_spark_row(row)

    @classmethod
    def from_spark_row(
        cls,
        row: "SparkRow",
        *,
        normalize: bool = True,
    ) -> "Response":
        def _to_python(value: Any) -> Any:
            if value is None:
                return None
            as_dict = getattr(value, "asDict", None)
            if callable(as_dict):
                return {str(k): _to_python(v) for k, v in value.asDict(recursive=True).items()}
            if isinstance(value, Mapping):
                return {str(k): _to_python(v) for k, v in value.items()}
            if isinstance(value, (list, tuple)):
                return [_to_python(v) for v in value]
            return value

        return cls.parse_mapping(_to_python(row), normalize=normalize)

    @classmethod
    def from_arrow_tabular(
        cls,
        batch: pa.RecordBatch | pa.Table | Iterator[pa.RecordBatch | pa.Table],
    ) -> Iterator["Response"]:
        def _iter_batches(
            obj: pa.RecordBatch | pa.Table | Iterator[pa.RecordBatch | pa.Table]
        ) -> Iterator[pa.RecordBatch]:
            if isinstance(obj, pa.RecordBatch):
                yield obj
            elif isinstance(obj, pa.Table):
                yield from obj.to_batches()
            else:
                for inner in obj:
                    yield from _iter_batches(inner)

        for rb in _iter_batches(batch):
            for info in rb.to_pylist():
                yield cls.parse_mapping(info, normalize=False)

    # ------------------------------------------------------------------
    # ASGI helpers
    # ------------------------------------------------------------------

    def _to_asgi_payload(self) -> tuple[bytes, dict[str, str], str]:
        body = self.buffer.to_bytes() if self.buffer is not None else b""
        media = self.media_type

        headers = {
            str(k): str(v)
            for k, v in (self.headers or {}).items()
            if str(k).lower() not in _HOP_BY_HOP
        }
        _pop_header(headers, "Content-Type")
        headers["Content-Length"] = str(len(body))

        if media.codec is not None and _parse_content_encoding(headers) is None:
            headers["Content-Encoding"] = media.codec.name

        return body, headers, media.mime_type.value

    def to_starlette(self) -> "StarletteResponse":
        from starlette.responses import Response as _StarletteResponse

        body, headers, media_type = self._to_asgi_payload()
        return _StarletteResponse(
            content=body,
            status_code=self.status_code,
            headers=headers,
            media_type=media_type,
        )

    def to_fastapi(self) -> "FastAPIResponse":
        try:
            from fastapi import Response as _FastAPIResponse
        except ImportError:
            return self.to_starlette()  # type: ignore[return-value]

        body, headers, media_type = self._to_asgi_payload()
        return _FastAPIResponse(
            content=body,
            status_code=self.status_code,
            headers=headers,
            media_type=media_type,
        )

    # ------------------------------------------------------------------
    # Functional helper
    # ------------------------------------------------------------------

    def apply(self, func: Callable[["Response"], "Response"]) -> "Response":
        return func(self)


"""Timezone normalisation and conversion utilities.

The :class:`Timezone` dataclass wraps an IANA timezone identifier (e.g.
``"Europe/Paris"``, ``"UTC"``) and provides:

* **Parsing** from strings — IANA names, common abbreviations (``CET``,
  ``EST``, …), and fixed UTC offsets (``+01:00``, ``UTC-05``).
* **Conversion helpers** — ``localize`` naive datetimes, ``convert`` between
  zones, extract ``utc_offset`` at a given instant.
* **Polars integration** — ``from_polars_dtype`` to extract the tz from a
  ``pl.Datetime``, ``polars_normalize`` to map a Series/Expr of timezone
  strings to canonical IANA names.
* **Pre-built constants** — ``Timezone.UTC``, ``Timezone.CET``, etc.
"""
from __future__ import annotations

import datetime as _dt
import re
from dataclasses import dataclass
from functools import lru_cache
from typing import Literal, Any, ClassVar, TYPE_CHECKING
from zoneinfo import ZoneInfo, available_timezones

if TYPE_CHECKING:
    import polars

__all__ = [
    "Timezone",
]

_TIMEZONE_ALIASES: dict[str, str] = {
    "UTC": "UTC",
    "ETC/UTC": "UTC",
    "+00:00": "UTC",
    "-00:00": "UTC",
    "GMT": "UTC",
    "Z": "UTC",
    "CET": "Europe/Paris",
    "CEST": "Europe/Paris",
    "WET": "Europe/Lisbon",
    "WEST": "Europe/Lisbon",
    "EET": "Europe/Helsinki",
    "EEST": "Europe/Helsinki",
    "ET": "America/New_York",
    "EST": "America/New_York",
    "EDT": "America/New_York",
    "CT": "America/Chicago",
    "CST": "America/Chicago",
    "CDT": "America/Chicago",
    "MT": "America/Denver",
    "MST": "America/Denver",
    "MDT": "America/Denver",
    "PT": "America/Los_Angeles",
    "PST": "America/Los_Angeles",
    "PDT": "America/Los_Angeles",
    "JST": "Asia/Tokyo",
    "KST": "Asia/Seoul",
    "HKT": "Asia/Hong_Kong",
    "SGT": "Asia/Singapore",
}

_OFFSET_RE = re.compile(r"^([+-])(\d{2}):?(\d{2})$")


@lru_cache(maxsize=1)
def _available_timezones_cached() -> frozenset[str]:
    return frozenset(available_timezones())


@dataclass(slots=True, frozen=True)
class Timezone:
    """An immutable wrapper around an IANA timezone identifier.

    Instances are created via :meth:`parse` (accepts strings, ``ZoneInfo``,
    other ``Timezone`` objects, or ``None`` → UTC) or directly::

        tz = Timezone("Europe/Paris")
        tz = Timezone.parse("CET")      # → Timezone("Europe/Paris")
        tz = Timezone.parse("+01:00")   # → Timezone("Etc/GMT-1")
    """

    iana: str

    # ── Pre-built constants (set after class body) ───────────────────────────
    UTC: ClassVar["Timezone"]
    CET: ClassVar["Timezone"]          # Europe/Paris (CET/CEST)
    WET: ClassVar["Timezone"]          # Europe/Lisbon
    EET: ClassVar["Timezone"]          # Europe/Helsinki
    EASTERN: ClassVar["Timezone"]      # America/New_York
    CENTRAL: ClassVar["Timezone"]      # America/Chicago
    MOUNTAIN: ClassVar["Timezone"]     # America/Denver
    PACIFIC: ClassVar["Timezone"]      # America/Los_Angeles
    JST: ClassVar["Timezone"]          # Asia/Tokyo
    SGT: ClassVar["Timezone"]          # Asia/Singapore

    # ── dunder ───────────────────────────────────────────────────────────────

    def __str__(self) -> str:
        return self.iana

    def __repr__(self) -> str:
        return f"Timezone({self.iana!r})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Timezone):
            return self.iana == other.iana
        if isinstance(other, str):
            return self.iana == other
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.iana)

    # ── Parsing ──────────────────────────────────────────────────────────────

    @classmethod
    def parse(cls, obj: Any) -> "Timezone":
        """Parse *obj* into a :class:`Timezone`.

        Accepted types:

        * ``Timezone`` — returned as-is.
        * ``str`` — delegates to :meth:`parse_str`.
        * ``ZoneInfo`` — extracts the ``key`` attribute.
        * ``None`` — returns :attr:`UTC`.

        Raises:
            TypeError: For unsupported types.
        """
        if isinstance(obj, cls):
            return obj
        if obj is None:
            return cls.UTC
        if isinstance(obj, ZoneInfo):
            return cls(obj.key)
        if isinstance(obj, str):
            return cls.parse_str(obj)
        raise TypeError(f"Cannot parse {type(obj).__name__} as {cls.__name__}")

    @classmethod
    def parse_str(cls, s: str) -> "Timezone":
        """Parse a timezone string.

        Resolution order:

        1. Exact IANA name (``"Europe/Paris"``).
        2. Known alias (``"CET"`` → ``Europe/Paris``).
        3. UTC offset (``"+01:00"``, ``"UTC-05"``, ``"-0530"``).

        Raises:
            TypeError: If *s* is not a ``str``.
            ValueError: If the string cannot be resolved.
        """
        if not isinstance(s, str):
            raise TypeError(f"Expected str, got {type(s).__name__}")

        raw = s.strip()
        if not raw:
            raise ValueError("Timezone string cannot be empty")

        # Exact IANA timezone
        if raw in _available_timezones_cached():
            return cls(raw)

        upper = raw.upper()

        # Known aliases
        if upper in _TIMEZONE_ALIASES:
            return cls(_TIMEZONE_ALIASES[upper])

        # Normalize "UTC+01:00" / "UTC-0500"
        offset_part = raw
        if upper.startswith("UTC") and len(raw) > 3:
            offset_part = raw[3:].strip()

        # Normalize "+01:00" / "-0500" to IANA Etc/GMT±X where possible
        m = _OFFSET_RE.match(offset_part)
        if m:
            sign, hh, mm = m.groups()
            hours = int(hh)
            minutes = int(mm)

            if minutes != 0:
                raise ValueError(
                    f"Cannot normalize non-hour-aligned UTC offset to IANA timezone: {s!r}"
                )

            if hours == 0:
                return cls("UTC")

            # IANA Etc/GMT sign convention is reversed
            etc_sign = "-" if sign == "+" else "+"
            return cls(f"Etc/GMT{etc_sign}{hours}")

        raise ValueError(f"Unknown timezone: {s!r}")

    # ── ZoneInfo interop ─────────────────────────────────────────────────────

    @lru_cache(maxsize=64)
    def to_zoneinfo(self) -> ZoneInfo:
        """Return the ``zoneinfo.ZoneInfo`` for this timezone."""
        return ZoneInfo(self.iana)

    @property
    def key(self) -> str:
        """Alias for :attr:`iana` — matches ``ZoneInfo.key``."""
        return self.iana

    # ── Offset / introspection ───────────────────────────────────────────────

    def utc_offset(self, at: _dt.datetime | None = None) -> _dt.timedelta:
        """Return the UTC offset at instant *at* (default: now).

        The result accounts for DST transitions.

        >>> Timezone("Europe/Paris").utc_offset(datetime(2024, 7, 1))
        datetime.timedelta(seconds=7200)  # +02:00 (CEST)
        """
        if at is None:
            at = _dt.datetime.now(_dt.timezone.utc)
        elif at.tzinfo is None:
            at = at.replace(tzinfo=_dt.timezone.utc)
        return at.astimezone(self.to_zoneinfo()).utcoffset() or _dt.timedelta(0)

    def utc_offset_hours(self, at: _dt.datetime | None = None) -> float:
        """Return the UTC offset in fractional hours.

        >>> Timezone("Asia/Kolkata").utc_offset_hours()
        5.5
        """
        return self.utc_offset(at).total_seconds() / 3600

    def is_utc(self) -> bool:
        """Return ``True`` if this timezone is UTC (or an equivalent)."""
        return self.iana in ("UTC", "Etc/UTC", "Etc/GMT", "Etc/GMT+0", "Etc/GMT-0", "GMT")

    def is_fixed_offset(self) -> bool:
        """Return ``True`` if this timezone has a fixed UTC offset (no DST)."""
        return self.iana.startswith("Etc/") or self.is_utc()

    def is_dst(self, at: _dt.datetime | None = None) -> bool:
        """Return ``True`` if DST is active at instant *at* (default: now).

        >>> Timezone("Europe/Paris").is_dst(datetime(2024, 7, 1))
        True
        >>> Timezone("Europe/Paris").is_dst(datetime(2024, 1, 1))
        False
        """
        if at is None:
            at = _dt.datetime.now(_dt.timezone.utc)
        elif at.tzinfo is None:
            at = at.replace(tzinfo=_dt.timezone.utc)
        localized = at.astimezone(self.to_zoneinfo())
        dst = localized.dst()
        return dst is not None and dst != _dt.timedelta(0)

    def dst_offset(self, at: _dt.datetime | None = None) -> _dt.timedelta:
        """Return the DST adjustment at instant *at*.

        Returns ``timedelta(0)`` when DST is not active or for fixed-offset
        timezones.
        """
        if at is None:
            at = _dt.datetime.now(_dt.timezone.utc)
        elif at.tzinfo is None:
            at = at.replace(tzinfo=_dt.timezone.utc)
        localized = at.astimezone(self.to_zoneinfo())
        return localized.dst() or _dt.timedelta(0)

    def abbreviation(self, at: _dt.datetime | None = None) -> str:
        """Return the timezone abbreviation (e.g. ``"CET"``, ``"CEST"``) at *at*.

        >>> Timezone("Europe/Paris").abbreviation(datetime(2024, 7, 1))
        'CEST'
        >>> Timezone("Europe/Paris").abbreviation(datetime(2024, 1, 1))
        'CET'
        """
        if at is None:
            at = _dt.datetime.now(_dt.timezone.utc)
        elif at.tzinfo is None:
            at = at.replace(tzinfo=_dt.timezone.utc)
        return at.astimezone(self.to_zoneinfo()).strftime("%Z")

    def distance_to(self, other: "Timezone", at: _dt.datetime | None = None) -> _dt.timedelta:
        """Return the offset difference between ``self`` and *other* at *at*.

        A positive result means *other* is ahead of ``self``.

        >>> Timezone.UTC.distance_to(Timezone.CET, datetime(2024, 7, 1))
        datetime.timedelta(seconds=7200)
        """
        return other.utc_offset(at) - self.utc_offset(at)

    @property
    def tzinfo(self) -> _dt.tzinfo:
        """Return as a stdlib ``tzinfo`` (same as ``to_zoneinfo()``)."""
        return self.to_zoneinfo()

    # ── Clock ────────────────────────────────────────────────────────────────

    def now(self) -> _dt.datetime:
        """Return the current wall-clock time in this timezone.

        The returned datetime is timezone-aware.
        """
        return _dt.datetime.now(self.to_zoneinfo())

    def today(self) -> _dt.date:
        """Return today's date in this timezone."""
        return self.now().date()

    # ── Conversion helpers ───────────────────────────────────────────────────

    def localize(self, naive: _dt.datetime) -> _dt.datetime:
        """Stamp a naive datetime with this timezone (no conversion).

        Raises:
            ValueError: If *naive* is already timezone-aware.
        """
        if naive.tzinfo is not None:
            raise ValueError(
                f"Cannot localize a timezone-aware datetime "
                f"(has tzinfo={naive.tzinfo!r}); use convert() instead"
            )
        return naive.replace(tzinfo=self.to_zoneinfo())

    def convert(self, aware: _dt.datetime) -> _dt.datetime:
        """Convert a timezone-aware datetime to this timezone.

        Raises:
            ValueError: If *aware* is naive.
        """
        if aware.tzinfo is None:
            raise ValueError(
                "Cannot convert a naive datetime; use localize() first"
            )
        return aware.astimezone(self.to_zoneinfo())

    def midnight(self, date: _dt.date | None = None) -> _dt.datetime:
        """Return midnight (00:00) in this timezone for *date* (default: today)."""
        if date is None:
            date = self.today()
        return _dt.datetime(date.year, date.month, date.day, tzinfo=self.to_zoneinfo())

    # ── Polars integration ───────────────────────────────────────────────────

    @classmethod
    def from_polars_dtype(cls, dtype: "polars.DataType") -> "Timezone | None":
        """Extract the timezone from a Polars ``Datetime`` dtype.

        Returns ``None`` if the dtype is not ``Datetime`` or has no timezone.

        >>> Timezone.from_polars_dtype(pl.Datetime("us", "Europe/Paris"))
        Timezone('Europe/Paris')
        """
        import polars as pl

        if isinstance(dtype, pl.Datetime) and dtype.time_zone:
            return cls.parse(dtype.time_zone)
        return None

    @classmethod
    def polars_normalize(
        cls,
        col: "polars.Series | polars.Expr",
        *,
        lazy: bool = True,
        return_value: Literal["iana"] = "iana",
    ) -> "polars.Series | polars.Expr":
        """Normalize timezone strings in a Polars column using alias replacement only."""
        import polars as pl

        if return_value != "iana":
            raise ValueError(f"Unsupported return_value: {return_value!r}")
        if not isinstance(col, (pl.Series, pl.Expr)):
            raise TypeError(f"Expected polars.Series | polars.Expr, got {type(col).__name__}")

        normalized = (
            col.cast(pl.Utf8)
            .str.strip_chars()
            .str.to_uppercase()
        )
        return normalized.replace_strict(_TIMEZONE_ALIASES, default=None)

    # ── Arrow integration ────────────────────────────────────────────────────

    @classmethod
    def from_arrow_type(cls, dtype: Any) -> "Timezone | None":
        """Extract the timezone from a PyArrow ``TimestampType``.

        Returns ``None`` if the type has no ``tz`` attribute or it is empty.

        >>> import pyarrow as pa
        >>> Timezone.from_arrow_type(pa.timestamp("us", tz="Europe/Paris"))
        Timezone('Europe/Paris')
        """
        tz = getattr(dtype, "tz", None)
        if tz:
            return cls.parse(tz)
        return None

    def arrow_timestamp_type(self, unit: str = "us") -> Any:
        """Return a ``pa.timestamp(unit, tz=self.iana)`` type.

        >>> Timezone.CET.arrow_timestamp_type("ns")
        TimestampType(timestamp[ns, tz=Europe/Paris])
        """
        import pyarrow as pa
        return pa.timestamp(unit, tz=self.iana)

    # ── Enumeration ──────────────────────────────────────────────────────────

    @classmethod
    def all_iana(cls) -> frozenset[str]:
        """Return all IANA timezone identifiers available on this system."""
        return _available_timezones_cached()


# ── Class-level constants (can't be set inside a frozen dataclass body) ──────
Timezone.UTC = Timezone("UTC")
Timezone.CET = Timezone("Europe/Paris")
Timezone.WET = Timezone("Europe/Lisbon")
Timezone.EET = Timezone("Europe/Helsinki")
Timezone.EASTERN = Timezone("America/New_York")
Timezone.CENTRAL = Timezone("America/Chicago")
Timezone.MOUNTAIN = Timezone("America/Denver")
Timezone.PACIFIC = Timezone("America/Los_Angeles")
Timezone.JST = Timezone("Asia/Tokyo")
Timezone.SGT = Timezone("Asia/Singapore")



from __future__ import annotations

import datetime as dt
import json
from dataclasses import dataclass
from urllib.parse import urlencode
from urllib.request import urlopen

from yggdrasil.data.enums.currency import Currency

__all__ = ["FxRate", "fetch_fx_rates"]


@dataclass(frozen=True, slots=True)
class FxRate:
    source: Currency
    target: Currency
    value: float


def _to_utc_datetime(value: dt.datetime) -> dt.datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=dt.timezone.utc)
    return value.astimezone(dt.timezone.utc)


def _fetch_daily_rates(source: Currency, target: Currency, start: dt.date, end: dt.date) -> dict[dt.date, float]:
    endpoint = f"https://api.frankfurter.app/{start.isoformat()}..{end.isoformat()}"
    qs = urlencode({"from": source.code, "to": target.code})
    with urlopen(f"{endpoint}?{qs}") as resp:  # nosec: B310 - trusted static host + encoded query
        payload = json.loads(resp.read().decode("utf-8"))

    rates = payload.get("rates", {})
    result: dict[dt.date, float] = {}
    for day, values in rates.items():
        raw = values.get(target.code)
        if raw is None:
            continue
        result[dt.date.fromisoformat(day)] = float(raw)
    if not result:
        raise ValueError(
            f"No FX rates returned for {source.code}/{target.code} between {start} and {end}"
        )
    return result


def fetch_fx_rates(
    source: Currency | str,
    target: Currency | str,
    start: dt.datetime,
    end: dt.datetime,
    *,
    sampling: int = 3600,
    as_polars: bool = True,
):
    src = Currency.parse(source)
    tgt = Currency.parse(target)

    start_utc = _to_utc_datetime(start)
    end_utc = _to_utc_datetime(end)

    if end_utc < start_utc:
        raise ValueError("end must be >= start")
    if sampling <= 0:
        raise ValueError("sampling must be > 0 seconds")

    daily = _fetch_daily_rates(src, tgt, start_utc.date(), end_utc.date())
    day_keys = sorted(daily)

    rows: list[tuple[dt.datetime, FxRate]] = []
    cursor = start_utc
    delta = dt.timedelta(seconds=int(sampling))

    while cursor <= end_utc:
        d = cursor.date()
        candidates = [day for day in day_keys if day <= d]
        use_day = candidates[-1] if candidates else day_keys[0]
        rate = FxRate(source=src, target=tgt, value=daily[use_day])
        rows.append((cursor, rate))
        cursor += delta

    if as_polars:
        from yggdrasil.polars.lib import polars as pl

        return pl.DataFrame(
            {
                "at": [at for at, _ in rows],
                "source": [r.source.code for _, r in rows],
                "target": [r.target.code for _, r in rows],
                "value": [r.value for _, r in rows],
            }
        )

    return [r for _, r in rows]



from __future__ import annotations

import gzip
import zlib
from typing import Final

from yggdrasil.pickle.ser.constants import (
    CODEC_GZIP,
    CODEC_NONE,
    CODEC_ZLIB,
    CODEC_ZSTD,
)
from yggdrasil.pickle.ser.errors import InvalidCodecError

try:
    import zstandard as _zstd
except Exception:  # pragma: no cover - optional dependency
    _zstd = None

__all__ = [
    "DEFAULT_CODEC",
    "codec_name",
    "default_codec",
    "compress_bytes",
    "decompress_bytes",
]

_CODEC_NAMES: Final[dict[int, str]] = {
    CODEC_NONE: "none",
    CODEC_GZIP: "gzip",
    CODEC_ZSTD: "zstd",
    CODEC_ZLIB: "zlib",
}

# Fast serializer-oriented defaults.
_GZIP_COMPRESSLEVEL: Final[int] = 1
_ZLIB_LEVEL: Final[int] = 1
_ZSTD_LEVEL: Final[int] = 1

_ZSTD_COMPRESSOR = _zstd.ZstdCompressor(level=_ZSTD_LEVEL) if _zstd is not None else None
_ZSTD_DECOMPRESSOR = _zstd.ZstdDecompressor() if _zstd is not None else None


def codec_name(codec: int) -> str:
    """Return a human-readable codec name."""
    try:
        return _CODEC_NAMES[codec]
    except KeyError as exc:
        raise InvalidCodecError(f"Unknown codec id: {codec}") from exc


def default_codec() -> int:
    """
    Return the preferred default compression codec for this runtime.

    Preference order:
    1. zstd, if installed
    2. zlib, always available in stdlib
    3. gzip, compatibility fallback
    """
    if _ZSTD_COMPRESSOR is not None:
        return CODEC_ZSTD
    return CODEC_ZLIB


DEFAULT_CODEC: Final[int] = default_codec()


def compress_bytes(data: bytes, codec: int) -> bytes:
    """Compress bytes using the configured wire codec."""
    if codec == CODEC_NONE:
        return data

    if codec == CODEC_GZIP:
        return gzip.compress(data, compresslevel=_GZIP_COMPRESSLEVEL)

    if codec == CODEC_ZLIB:
        return zlib.compress(data, level=_ZLIB_LEVEL)

    if codec == CODEC_ZSTD:
        if _ZSTD_COMPRESSOR is None:
            raise InvalidCodecError(
                "CODEC_ZSTD requested but zstandard is not installed"
            )
        return _ZSTD_COMPRESSOR.compress(data)

    raise InvalidCodecError(f"Unknown codec id: {codec}")


def decompress_bytes(data: bytes, codec: int) -> bytes:
    """Decompress bytes using the configured wire codec."""
    if codec == CODEC_NONE:
        return data

    if codec == CODEC_GZIP:
        return gzip.decompress(data)

    if codec == CODEC_ZLIB:
        return zlib.decompress(data)

    if codec == CODEC_ZSTD:
        if _ZSTD_DECOMPRESSOR is None:
            raise InvalidCodecError(
                "CODEC_ZSTD encountered but zstandard is not installed"
            )
        return _ZSTD_DECOMPRESSOR.decompress(data)

    raise InvalidCodecError(f"Unknown codec id: {codec}")


"""Geographic feature zones: mountain ranges, massifs, passes, and rivers.

Coordinates are representative points (approximate centroid or highest-peak
location) in WGS-84 (EPSG:4326).  All zones use ``GeoZoneType.ZONE``.

Sources
-------
* Wikipedia – individual mountain / range / pass / river articles
* OpenStreetMap nominatim – cross-checked centroids
"""
from __future__ import annotations

from typing import Optional

from .builders import _validate_unique_attrs, _from_coordinates_with_optional_metadata
from .geozone import GeoZone, GeoZoneType

__all__ = ["load_mountains"]

_SRC = "seed: module (geographic feature representative point; Wikipedia/OSM cross-check)"


def _feature(
    key: str,
    name: str,
    lat: float,
    lon: float,
    *,
    aliases: tuple[str, ...] = (),
    country_iso: Optional[str] = None,
    country_name: Optional[str] = None,
    tz: Optional[str] = None,
    confidence: str = "medium",
    coord_kind: str = "representative_point",
) -> GeoZone:
    """Register a geographic feature as a ``GeoZoneType.ZONE``."""
    return GeoZone.put(
        _from_coordinates_with_optional_metadata(
            gtype=GeoZoneType.ZONE,
            lat=lat,
            lon=lon,
            key=key,
            aliases=aliases,
            name=name,
            country_iso=country_iso,
            country_name=country_name,
            tz=tz,
            ccy=None,
            coord_source=_SRC,
            coord_kind=coord_kind,
            confidence=confidence,
        )
    )


# ---------------------------------------------------------------------------
# Table columns: attr, key, name, lat, lon, kwargs
# ---------------------------------------------------------------------------

# ── Mountain ranges ─────────────────────────────────────────────────────────

MOUNTAIN_RANGES = [
    # Alps – the main arc (CH/AT/IT/FR/DE/LI/SI)
    ("ALPS", "ALPS", "Alps",
     46.5000, 10.5000,
     {"aliases": ("THE_ALPS", "ALPINE"), "confidence": "high"}),

    # Western Alps
    ("WESTERN_ALPS", "WESTERN_ALPS", "Western Alps",
     44.8000, 7.0000,
     {"aliases": ("ALPS_WEST",)}),

    # Central Alps
    ("CENTRAL_ALPS", "CENTRAL_ALPS", "Central Alps",
     46.6000, 9.5000,
     {"aliases": ("ALPS_CENTRAL",)}),

    # Eastern Alps
    ("EASTERN_ALPS", "EASTERN_ALPS", "Eastern Alps",
     47.2000, 13.5000,
     {"aliases": ("ALPS_EAST",)}),

    # Swiss Alps (centroid of the Swiss portion)
    ("SWISS_ALPS", "SWISS_ALPS", "Swiss Alps",
     46.5500, 8.1000,
     {"aliases": ("SCHWEIZER_ALPEN",), "country_iso": "CH", "country_name": "Switzerland",
      "tz": "Europe/Zurich"}),

    # French Alps
    ("FRENCH_ALPS", "FRENCH_ALPS", "French Alps",
     45.3000, 6.4000,
     {"aliases": ("ALPES_FRANCAISES",), "country_iso": "FR", "country_name": "France",
      "tz": "Europe/Paris"}),

    # Italian Alps
    ("ITALIAN_ALPS", "ITALIAN_ALPS", "Italian Alps",
     46.0000, 11.5000,
     {"aliases": ("ALPI_ITALIANE",), "country_iso": "IT", "country_name": "Italy",
      "tz": "Europe/Rome"}),

    # Austrian Alps
    ("AUSTRIAN_ALPS", "AUSTRIAN_ALPS", "Austrian Alps",
     47.2000, 13.0000,
     {"aliases": ("OESTERREICHISCHE_ALPEN",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    # Pennine Alps (CH/IT; highest massif of the Alps)
    ("PENNINE_ALPS", "PENNINE_ALPS", "Pennine Alps",
     45.9766, 7.6586,
     {"aliases": ("VALAIS_ALPS", "WALLIS_ALPEN")}),

    # Bernese Alps (CH)
    ("BERNESE_ALPS", "BERNESE_ALPS", "Bernese Alps",
     46.5782, 7.9800,
     {"aliases": ("BERNESE_OBERLAND", "BERNER_ALPEN"), "country_iso": "CH",
      "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lepontine Alps (CH/IT)
    ("LEPONTINE_ALPS", "LEPONTINE_ALPS", "Lepontine Alps",
     46.4500, 8.9000,
     {"aliases": ("LEPONTIC_ALPS",)}),

    # Graian Alps (FR/IT)
    ("GRAIAN_ALPS", "GRAIAN_ALPS", "Graian Alps",
     45.5000, 7.1000,
     {"aliases": ("ALPES_GRAIENNES",)}),

    # Cottian Alps (FR/IT)
    ("COTTIAN_ALPS", "COTTIAN_ALPS", "Cottian Alps",
     44.6500, 6.8000,
     {"aliases": ("ALPES_COTTIENNES",)}),

    # Maritime Alps (FR/IT)
    ("MARITIME_ALPS", "MARITIME_ALPS", "Maritime Alps",
     44.1500, 7.1000,
     {"aliases": ("ALPES_MARITIMES_RANGE",)}),

    # Rhaetian Alps (CH/AT/IT)
    ("RHAETIAN_ALPS", "RHAETIAN_ALPS", "Rhaetian Alps",
     46.5000, 10.0000,
     {"aliases": ("RAETISCHE_ALPEN",)}),

    # Glarus Alps (CH)
    ("GLARUS_ALPS", "GLARUS_ALPS", "Glarus Alps",
     46.9000, 9.0000,
     {"country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Ötztaler Alps (AT/IT)
    ("OTZTAL_ALPS", "OTZTAL_ALPS", "Ötztal Alps",
     46.8300, 10.8500,
     {"aliases": ("OTZTALER_ALPEN",)}),

    # Stubai Alps (AT)
    ("STUBAI_ALPS", "STUBAI_ALPS", "Stubai Alps",
     47.0000, 11.2000,
     {"aliases": ("STUBAIER_ALPEN",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    # Tauern (High Tauern + Low Tauern, AT)
    ("HIGH_TAUERN", "HIGH_TAUERN", "High Tauern",
     47.0742, 12.6950,
     {"aliases": ("HOHE_TAUERN",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    ("LOW_TAUERN", "LOW_TAUERN", "Low Tauern",
     47.3000, 14.2000,
     {"aliases": ("NIEDERE_TAUERN",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    # Dolomites (IT)
    ("DOLOMITES", "DOLOMITES", "Dolomites",
     46.4102, 11.8440,
     {"aliases": ("DOLOMITI",), "country_iso": "IT", "country_name": "Italy",
      "tz": "Europe/Rome", "confidence": "high"}),

    # Apennines (IT)
    ("APENNINES", "APENNINES", "Apennine Mountains",
     43.1500, 13.1000,
     {"aliases": ("APENNINI",), "country_iso": "IT", "country_name": "Italy",
      "tz": "Europe/Rome"}),

    # Pyrenees (FR/ES/AD)
    ("PYRENEES", "PYRENEES", "Pyrenees",
     42.7500, 1.0000,
     {"aliases": ("PIRINEOS", "PIRENEES"), "confidence": "high"}),

    # Carpathians (PL/SK/CZ/UA/RO/RS/AT)
    ("CARPATHIANS", "CARPATHIANS", "Carpathian Mountains",
     49.0000, 23.0000,
     {"aliases": ("CARPATES", "KARPATEN"), "confidence": "high"}),

    ("TATRA", "TATRA", "Tatra Mountains",
     49.2292, 19.9808,
     {"aliases": ("TATRY", "HIGH_TATRA")}),

    # Scandinavian Mountains
    ("SCANDINAVIAN_MOUNTAINS", "SCANDINAVIAN_MOUNTAINS", "Scandinavian Mountains",
     63.0000, 13.5000,
     {"aliases": ("SCANDES", "KJOLEN")}),

    # Black Forest (DE)
    ("BLACK_FOREST", "BLACK_FOREST", "Black Forest",
     48.2000, 8.2000,
     {"aliases": ("SCHWARZWALD",), "country_iso": "DE", "country_name": "Germany",
      "tz": "Europe/Berlin"}),

    # Vosges (FR)
    ("VOSGES", "VOSGES", "Vosges Mountains",
     48.1500, 7.1000,
     {"country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Jura (FR/CH)
    ("JURA", "JURA", "Jura Mountains",
     47.0000, 6.5000,
     {"aliases": ("JURA_MOUNTAINS",)}),

    # Massif Central (FR)
    ("MASSIF_CENTRAL", "MASSIF_CENTRAL", "Massif Central",
     45.5000, 3.0000,
     {"country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Caucasus
    ("CAUCASUS", "CAUCASUS", "Caucasus Mountains",
     42.6000, 44.0000,
     {"aliases": ("KAVKAZ",), "confidence": "high"}),

    ("GREATER_CAUCASUS", "GREATER_CAUCASUS", "Greater Caucasus",
     43.1000, 44.0000,
     {"aliases": ("BOLSHOY_KAVKAZ",)}),

    ("LESSER_CAUCASUS", "LESSER_CAUCASUS", "Lesser Caucasus",
     40.5000, 44.5000,
     {"aliases": ("MALYI_KAVKAZ",)}),

    # Ural (RU)
    ("URAL", "URAL", "Ural Mountains",
     60.0000, 59.5000,
     {"aliases": ("URALS",), "country_iso": "RU", "country_name": "Russia"}),

    # Balkans (BGR/SRB/MK)
    ("BALKAN_MOUNTAINS", "BALKAN_MOUNTAINS", "Balkan Mountains",
     42.8500, 25.3000,
     {"aliases": ("STARA_PLANINA",)}),

    # Rhodopes (BG/GR)
    ("RHODOPES", "RHODOPES", "Rhodope Mountains",
     41.8000, 24.5000,
     {"aliases": ("RHODOPI",)}),

    # Atlas (MA/DZ/TN)
    ("ATLAS", "ATLAS", "Atlas Mountains",
     31.0000, -5.0000,
     {"aliases": ("ATLAS_MOUNTAINS",), "confidence": "high"}),

    # Himalayas
    ("HIMALAYAS", "HIMALAYAS", "Himalayas",
     28.0000, 84.0000,
     {"aliases": ("HIMALAYA",), "confidence": "high"}),

    # Andes
    ("ANDES", "ANDES", "Andes",
     -20.0000, -68.0000,
     {"aliases": ("CORDILLERA_DE_LOS_ANDES",), "confidence": "high"}),

    # Rocky Mountains
    ("ROCKY_MOUNTAINS", "ROCKY_MOUNTAINS", "Rocky Mountains",
     43.0000, -110.0000,
     {"aliases": ("ROCKIES",), "country_iso": "US", "country_name": "United States",
      "tz": "America/Denver", "confidence": "high"}),
]


# ── Individual peaks ─────────────────────────────────────────────────────────

PEAKS = [
    # Mont Blanc – highest peak in the Alps and Western Europe (FR/IT border)
    ("MONT_BLANC", "MONT_BLANC", "Mont Blanc",
     45.8326, 6.8652,
     {"aliases": ("MONTE_BIANCO",), "coord_kind": "summit", "confidence": "high"}),

    # Monte Rosa – highest entirely in Switzerland
    ("MONTE_ROSA", "MONTE_ROSA", "Monte Rosa",
     45.9374, 7.8679,
     {"aliases": ("DUFOURSPITZE",), "coord_kind": "summit", "confidence": "high"}),

    # Matterhorn (CH/IT)
    ("MATTERHORN", "MATTERHORN", "Matterhorn",
     45.9766, 7.6586,
     {"aliases": ("CERVINO", "CERVIN"), "coord_kind": "summit", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Jungfrau (CH)
    ("JUNGFRAU", "JUNGFRAU", "Jungfrau",
     46.5371, 7.9626,
     {"coord_kind": "summit", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Eiger (CH)
    ("EIGER", "EIGER", "Eiger",
     46.5775, 8.0054,
     {"coord_kind": "summit", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Mönch (CH)
    ("MONCH", "MONCH", "Mönch",
     46.5589, 7.9986,
     {"aliases": ("MOENCH",), "coord_kind": "summit", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Piz Bernina – highest peak in the Rhaetian Alps (CH)
    ("PIZ_BERNINA", "PIZ_BERNINA", "Piz Bernina",
     46.3783, 9.9088,
     {"aliases": ("BERNINA_PEAK",), "coord_kind": "summit", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Grossglockner – highest peak in Austria
    ("GROSSGLOCKNER", "GROSSGLOCKNER", "Grossglockner",
     47.0742, 12.6950,
     {"aliases": ("GROSS_GLOCKNER",), "coord_kind": "summit", "confidence": "high",
      "country_iso": "AT", "country_name": "Austria", "tz": "Europe/Vienna"}),

    # Gran Paradiso – highest entirely in Italy
    ("GRAN_PARADISO", "GRAN_PARADISO", "Gran Paradiso",
     45.5173, 7.2699,
     {"coord_kind": "summit", "confidence": "high",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome"}),

    # Ortler (IT)
    ("ORTLER", "ORTLER", "Ortler",
     46.5075, 10.5436,
     {"aliases": ("ORTLES",), "coord_kind": "summit",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome"}),

    # Zugspitze – highest peak in Germany
    ("ZUGSPITZE", "ZUGSPITZE", "Zugspitze",
     47.4211, 10.9850,
     {"coord_kind": "summit", "confidence": "high",
      "country_iso": "DE", "country_name": "Germany", "tz": "Europe/Berlin"}),

    # Pic du Midi d'Ossau (FR/ES, Pyrenees)
    ("PIC_DU_MIDI", "PIC_DU_MIDI", "Pic du Midi d'Ossau",
     42.8417, -0.4386,
     {"aliases": ("PIC_DU_MIDI_OSSAU",), "coord_kind": "summit",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Elbrus – highest peak in Europe (RU, Caucasus)
    ("ELBRUS", "ELBRUS", "Mount Elbrus",
     43.3499, 42.4453,
     {"aliases": ("MOUNT_ELBRUS",), "coord_kind": "summit", "confidence": "high",
      "country_iso": "RU", "country_name": "Russia"}),

    # Everest
    ("EVEREST", "EVEREST", "Mount Everest",
     27.9881, 86.9250,
     {"aliases": ("MT_EVEREST", "SAGARMATHA", "CHOMOLUNGMA"),
      "coord_kind": "summit", "confidence": "high"}),

    # Aconcagua
    ("ACONCAGUA", "ACONCAGUA", "Aconcagua",
     -32.6532, -70.0109,
     {"coord_kind": "summit", "confidence": "high",
      "country_iso": "AR", "country_name": "Argentina"}),
]


# ── Alpine passes ─────────────────────────────────────────────────────────────

PASSES = [
    # Great St Bernard Pass (CH/IT)
    ("GREAT_ST_BERNARD", "GREAT_ST_BERNARD", "Great St Bernard Pass",
     45.8695, 7.1706,
     {"aliases": ("GRAN_SAN_BERNARDO", "GRAND_SAINT_BERNARD"),
      "coord_kind": "pass", "confidence": "high"}),

    # Little St Bernard Pass (FR/IT)
    ("LITTLE_ST_BERNARD", "LITTLE_ST_BERNARD", "Little St Bernard Pass",
     45.6847, 6.8844,
     {"aliases": ("PETIT_SAINT_BERNARD",), "coord_kind": "pass"}),

    # Simplon Pass (CH/IT)
    ("SIMPLON", "SIMPLON", "Simplon Pass",
     46.2535, 8.0331,
     {"aliases": ("SEMPIONE", "SIMPLON_PASS"), "coord_kind": "pass", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Gotthard Pass (CH)
    ("GOTTHARD", "GOTTHARD", "Gotthard Pass",
     46.5568, 8.5647,
     {"aliases": ("ST_GOTTHARD", "SAN_GOTTARDO", "SAINT_GOTTHARD"),
      "coord_kind": "pass", "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Furka Pass (CH)
    ("FURKA", "FURKA", "Furka Pass",
     46.5736, 8.4154,
     {"aliases": ("FURKA_PASS",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Grimsel Pass (CH)
    ("GRIMSEL", "GRIMSEL", "Grimsel Pass",
     46.5609, 8.3350,
     {"aliases": ("GRIMSEL_PASS",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Susten Pass (CH)
    ("SUSTEN", "SUSTEN", "Susten Pass",
     46.7287, 8.4523,
     {"aliases": ("SUSTEN_PASS",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Nufenen Pass (CH)
    ("NUFENEN", "NUFENEN", "Nufenen Pass",
     46.4786, 8.3836,
     {"aliases": ("NOVENA",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Maloja Pass (CH)
    ("MALOJA", "MALOJA", "Maloja Pass",
     46.4026, 9.6956,
     {"aliases": ("MALOJA_PASS",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Bernina Pass (CH)
    ("BERNINA", "BERNINA", "Bernina Pass",
     46.4118, 10.0297,
     {"aliases": ("BERNINA_PASS",), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Julier Pass (CH)
    ("JULIER", "JULIER", "Julier Pass",
     46.4722, 9.7354,
     {"aliases": ("JULIER_PASS", "PASSO_DEL_GIULIA"), "coord_kind": "pass",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Splügen Pass (CH/IT)
    ("SPLUGEN", "SPLUGEN", "Splügen Pass",
     46.5044, 9.3272,
     {"aliases": ("SPLUGEN_PASS", "PASSO_DELLO_SPLUGA"), "coord_kind": "pass"}),

    # Brenner Pass (AT/IT) – major trans-Alpine route
    ("BRENNER", "BRENNER", "Brenner Pass",
     47.0067, 11.5069,
     {"aliases": ("BRENNER_PASS", "PASSO_DEL_BRENNERO"),
      "coord_kind": "pass", "confidence": "high"}),

    # Reschen Pass (AT/IT)
    ("RESCHEN", "RESCHEN", "Reschen Pass",
     46.8517, 10.5358,
     {"aliases": ("PASSO_DI_RESIA",), "coord_kind": "pass"}),

    # Stelvio Pass (IT) – highest paved road in the Alps (2757 m)
    ("STELVIO", "STELVIO", "Stelvio Pass",
     46.5284, 10.4537,
     {"aliases": ("PASSO_DELLO_STELVIO",), "coord_kind": "pass",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome"}),

    # Col du Mont Cenis (FR/IT)
    ("MONT_CENIS", "MONT_CENIS", "Col du Mont Cenis",
     45.2614, 6.9028,
     {"aliases": ("COL_DU_MONT_CENIS",), "coord_kind": "pass"}),

    # Col de la Faucille (FR)
    ("FAUCILLE", "FAUCILLE", "Col de la Faucille",
     46.3667, 6.0269,
     {"aliases": ("COL_DE_LA_FAUCILLE",), "coord_kind": "pass",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),
]


# ── Rivers ───────────────────────────────────────────────────────────────────

RIVERS = [
    # Rhone – source in Valais (CH), flows through Lake Geneva, south through France to Med
    ("RHONE", "RHONE", "Rhône",
     45.0000, 5.5000,
     {"aliases": ("RHONE_RIVER", "FLEUVE_RHONE"),
      "coord_kind": "river_centroid", "confidence": "high"}),

    # Rhone source (Rhône Glacier, CH)
    ("RHONE_SOURCE", "RHONE_SOURCE", "Rhône Source",
     46.5733, 8.3733,
     {"aliases": ("RHONE_GLACIER",), "coord_kind": "source",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Rhine – source in CH (Graubünden), flows via Germany/France to North Sea
    ("RHINE", "RHINE", "Rhine",
     49.0000, 8.0000,
     {"aliases": ("RHEIN", "RHIN", "RHINE_RIVER"),
      "coord_kind": "river_centroid", "confidence": "high"}),

    # Rhine source (Lake Toma, CH)
    ("RHINE_SOURCE", "RHINE_SOURCE", "Rhine Source",
     46.6389, 8.6833,
     {"aliases": ("LAKE_TOMA",), "coord_kind": "source",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Aare (CH)
    ("AARE", "AARE", "Aare",
     46.8000, 7.8000,
     {"aliases": ("AAR",), "coord_kind": "river_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Inn (CH/AT/DE) – tributary of the Danube
    ("INN", "INN", "Inn",
     47.0000, 10.9000,
     {"aliases": ("INN_RIVER",), "coord_kind": "river_centroid"}),

    # Danube – longest EU river
    ("DANUBE", "DANUBE", "Danube",
     47.0000, 20.0000,
     {"aliases": ("DONAU", "DUNA", "DUNAV", "DUNAREA"),
      "coord_kind": "river_centroid", "confidence": "high"}),

    # Po (IT) – longest river entirely in Italy
    ("PO", "PO", "Po",
     45.0000, 10.5000,
     {"aliases": ("PO_RIVER",), "coord_kind": "river_centroid",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome"}),

    # Adige (IT)
    ("ADIGE", "ADIGE", "Adige",
     46.0000, 11.3000,
     {"aliases": ("ETSCH",), "coord_kind": "river_centroid",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome"}),

    # Isère (FR) – Alpine tributary of the Rhône
    ("ISERE", "ISERE", "Isère",
     45.2000, 5.7000,
     {"aliases": ("ISERE_RIVER",), "coord_kind": "river_centroid",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Durance (FR) – right-bank tributary of the Rhône
    ("DURANCE", "DURANCE", "Durance",
     43.9000, 5.4000,
     {"coord_kind": "river_centroid",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Saône (FR) – main tributary of the Rhône
    ("SAONE", "SAONE", "Saône",
     46.5000, 5.0000,
     {"aliases": ("SAONE_RIVER",), "coord_kind": "river_centroid",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Ticino (CH/IT)
    ("TICINO", "TICINO", "Ticino",
     46.0000, 8.9000,
     {"aliases": ("TESSIN",), "coord_kind": "river_centroid"}),

    # Reuss (CH) – drains Gotthard area into Lake Lucerne → Aare
    ("REUSS", "REUSS", "Reuss",
     47.0000, 8.3000,
     {"coord_kind": "river_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Valais / Vispa (CH) – drains Matterhorn/Monte Rosa area into Rhône
    ("VISPA", "VISPA", "Vispa",
     46.1200, 7.8000,
     {"aliases": ("VISP", "VIEGE"), "coord_kind": "river_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lütschine (CH) – drains Jungfrau area
    ("LUTSCHINE", "LUTSCHINE", "Lütschine",
     46.6500, 7.9000,
     {"aliases": ("LUTSCHINE",), "coord_kind": "river_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),
]


# ── Alpine lakes ──────────────────────────────────────────────────────────────

LAKES = [
    # Lake Geneva / Lac Léman (CH/FR)
    ("LAKE_GENEVA", "LAKE_GENEVA", "Lake Geneva",
     46.4500, 6.5833,
     {"aliases": ("LAC_LEMAN", "LEMAN", "GENFERSEE"),
      "coord_kind": "lake_centroid", "confidence": "high"}),

    # Lake Constance (DE/AT/CH)
    ("LAKE_CONSTANCE", "LAKE_CONSTANCE", "Lake Constance",
     47.6500, 9.3000,
     {"aliases": ("BODENSEE",), "coord_kind": "lake_centroid", "confidence": "high"}),

    # Lake Lucerne (CH)
    ("LAKE_LUCERNE", "LAKE_LUCERNE", "Lake Lucerne",
     47.0000, 8.4333,
     {"aliases": ("VIERWALDSTATTERSEE",), "coord_kind": "lake_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lake Zurich (CH)
    ("LAKE_ZURICH", "LAKE_ZURICH", "Lake Zurich",
     47.2394, 8.7184,
     {"aliases": ("ZUERICHSEE",), "coord_kind": "lake_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lake Maggiore (IT/CH)
    ("LAKE_MAGGIORE", "LAKE_MAGGIORE", "Lake Maggiore",
     45.8908, 8.5726,
     {"aliases": ("LAGO_MAGGIORE", "VERBANO"), "coord_kind": "lake_centroid",
      "confidence": "high"}),

    # Lake Como (IT)
    ("LAKE_COMO", "LAKE_COMO", "Lake Como",
     46.0069, 9.2685,
     {"aliases": ("LAGO_DI_COMO", "LARIO"), "coord_kind": "lake_centroid",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome",
      "confidence": "high"}),

    # Lake Garda (IT) – largest lake in Italy
    ("LAKE_GARDA", "LAKE_GARDA", "Lake Garda",
     45.6389, 10.6556,
     {"aliases": ("LAGO_DI_GARDA", "BENACO"), "coord_kind": "lake_centroid",
      "country_iso": "IT", "country_name": "Italy", "tz": "Europe/Rome",
      "confidence": "high"}),

    # Lake Brienz (CH)
    ("LAKE_BRIENZ", "LAKE_BRIENZ", "Lake Brienz",
     46.7200, 7.9700,
     {"aliases": ("BRIENZERSEE",), "coord_kind": "lake_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lake Thun (CH)
    ("LAKE_THUN", "LAKE_THUN", "Lake Thun",
     46.6900, 7.7400,
     {"aliases": ("THUNERSEE",), "coord_kind": "lake_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Lake Lugano (CH/IT)
    ("LAKE_LUGANO", "LAKE_LUGANO", "Lake Lugano",
     45.9667, 8.9333,
     {"aliases": ("LAGO_DI_LUGANO", "CERESIO"), "coord_kind": "lake_centroid"}),

    # Lake Neuchâtel (CH)
    ("LAKE_NEUCHATEL", "LAKE_NEUCHATEL", "Lake Neuchâtel",
     46.9667, 6.9167,
     {"aliases": ("NEUENBURGERSEE",), "coord_kind": "lake_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),
]


# ── Glaciers ──────────────────────────────────────────────────────────────────

GLACIERS = [
    # Aletsch – largest glacier in the Alps (CH)
    ("ALETSCH", "ALETSCH", "Aletsch Glacier",
     46.5083, 8.0361,
     {"aliases": ("GROSSER_ALETSCHGLETSCHER",), "coord_kind": "glacier_centroid",
      "confidence": "high",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Rhône Glacier (CH) – source of the Rhône
    ("RHONE_GLACIER", "RHONE_GLACIER", "Rhône Glacier",
     46.5733, 8.3733,
     {"aliases": ("RHONEGLETSCHER",), "coord_kind": "glacier_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Gorner Glacier (CH) – near Zermatt, second-largest in the Alps
    ("GORNER", "GORNER", "Gorner Glacier",
     45.9800, 7.7900,
     {"aliases": ("GORNERGLETSCHER",), "coord_kind": "glacier_centroid",
      "country_iso": "CH", "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Mer de Glace (FR) – largest glacier in France
    ("MER_DE_GLACE", "MER_DE_GLACE", "Mer de Glace",
     45.8900, 6.9300,
     {"coord_kind": "glacier_centroid",
      "country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),

    # Pasterze (AT) – longest glacier in the Eastern Alps
    ("PASTERZE", "PASTERZE", "Pasterze Glacier",
     47.0742, 12.6500,
     {"coord_kind": "glacier_centroid",
      "country_iso": "AT", "country_name": "Austria", "tz": "Europe/Vienna"}),
]


# ── Alpine valleys & regions ──────────────────────────────────────────────────

VALLEYS = [
    # Valais / Wallis (CH) – canton of Switzerland, home of Rhône and Matterhorn
    ("VALAIS", "VALAIS", "Valais",
     46.2000, 7.5000,
     {"aliases": ("WALLIS",), "country_iso": "CH", "country_name": "Switzerland",
      "tz": "Europe/Zurich", "confidence": "high"}),

    # Rhône Valley (CH) – Haut-Valais portion
    ("RHONE_VALLEY", "RHONE_VALLEY", "Rhône Valley",
     46.3000, 7.8000,
     {"aliases": ("RHONETAL",), "country_iso": "CH", "country_name": "Switzerland",
      "tz": "Europe/Zurich"}),

    # Engadin / Engadine (CH) – high Alpine valley, Inn headwaters
    ("ENGADIN", "ENGADIN", "Engadin",
     46.5000, 9.8500,
     {"aliases": ("ENGADINE",), "country_iso": "CH", "country_name": "Switzerland",
      "tz": "Europe/Zurich", "confidence": "high"}),

    # Ticino (CH) – canton south of Gotthard
    ("CANTON_TICINO", "CANTON_TICINO", "Canton Ticino",
     46.1500, 8.9000,
     {"aliases": ("TESSIN",), "country_iso": "CH", "country_name": "Switzerland",
      "tz": "Europe/Zurich"}),

    # Graubünden (CH) – largest canton, Rhine/Inn source
    ("GRAUBUENDEN", "GRAUBUENDEN", "Graubünden",
     46.6500, 9.5000,
     {"aliases": ("GRISONS", "GRIGIONI"), "country_iso": "CH",
      "country_name": "Switzerland", "tz": "Europe/Zurich"}),

    # Savoy (FR) – Alpine region
    ("SAVOIE", "SAVOIE", "Savoie",
     45.5667, 6.5000,
     {"aliases": ("SAVOY", "HAUTE_SAVOIE"), "country_iso": "FR",
      "country_name": "France", "tz": "Europe/Paris"}),

    # Valle d'Aosta (IT) – autonomous region, Mont Blanc side
    ("VALLE_DAOSTA", "VALLE_DAOSTA", "Valle d'Aosta",
     45.7375, 7.4262,
     {"aliases": ("AOSTA_VALLEY",), "country_iso": "IT",
      "country_name": "Italy", "tz": "Europe/Rome"}),

    # South Tyrol / Alto Adige (IT)
    ("SOUTH_TYROL", "SOUTH_TYROL", "South Tyrol",
     46.7000, 11.3000,
     {"aliases": ("ALTO_ADIGE", "SUEDTIROL"), "country_iso": "IT",
      "country_name": "Italy", "tz": "Europe/Rome"}),

    # Tyrol (AT)
    ("TYROL", "TYROL", "Tyrol",
     47.2500, 11.4000,
     {"aliases": ("TIROL",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    # Vorarlberg (AT)
    ("VORARLBERG", "VORARLBERG", "Vorarlberg",
     47.2500, 9.9000,
     {"country_iso": "AT", "country_name": "Austria", "tz": "Europe/Vienna"}),

    # Carinthia (AT)
    ("CARINTHIA", "CARINTHIA", "Carinthia",
     46.7000, 13.9000,
     {"aliases": ("KAERNTEN",), "country_iso": "AT", "country_name": "Austria",
      "tz": "Europe/Vienna"}),

    # Dauphiné (FR)
    ("DAUPHINE", "DAUPHINE", "Dauphiné",
     45.0000, 5.7000,
     {"country_iso": "FR", "country_name": "France", "tz": "Europe/Paris"}),
]


# ---------------------------------------------------------------------------
# Loader
# ---------------------------------------------------------------------------

def load_mountains() -> None:
    all_tables = [
        (MOUNTAIN_RANGES, "MOUNTAIN_RANGES"),
        (PEAKS, "PEAKS"),
        (PASSES, "PASSES"),
        (RIVERS, "RIVERS"),
        (LAKES, "LAKES"),
        (GLACIERS, "GLACIERS"),
        (VALLEYS, "VALLEYS"),
    ]

    for table, label in all_tables:
        _validate_unique_attrs(table, label)

    for table, _label in all_tables:
        for attr, key, name, lat, lon, kwargs in table:
            setattr(
                GeoZone,
                attr,
                _feature(key, name, lat, lon, **kwargs),
            )




from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from io import BytesIO as StdBytesIO
from typing import ClassVar, Generic, TypeVar

from yggdrasil.polars.lib import polars as pl
import pyarrow as pa

from yggdrasil.io import BytesIO
from yggdrasil.pickle.ser.serialized import Serialized
from yggdrasil.pickle.ser.tags import Tags

from .pyarrow import (
    _merge_metadata,
    _table_from_ipc_file_buffer,
    _table_to_ipc_file_buffer,
)

__all__ = [
    "TPolars",
    "PolarsSerialized",
    "PolarsDataFrameSerialized",
    "PolarsSeriesSerialized",
    "PolarsLazyFrameSerialized",
    "PolarsExprSerialized",
    "PolarsSchemaSerialized",
    "PolarsDataTypeSerialized",
]

TPolars = TypeVar("TPolars", bound=object)

_SENTINEL_FIELD_NAME = "__ygg_value__"


def _dataframe_to_arrow_table(df: pl.DataFrame) -> pa.Table:
    return df.to_arrow()


def _series_to_arrow_table(series: pl.Series) -> pa.Table:
    name = series.name if series.name is not None else _SENTINEL_FIELD_NAME
    return series.to_frame(name=name).to_arrow()


def _arrow_table_to_dataframe(table: pa.Table) -> pl.DataFrame:
    return pl.from_arrow(table)


def _arrow_table_to_series(table: pa.Table) -> pl.Series:
    df = pl.from_arrow(table)
    if df.width != 1:
        raise ValueError(
            f"POLARS_SERIES payload must contain exactly 1 column, got {df.width}"
        )
    return df.to_series(0)


def _simple_dtype_name(dtype: object) -> str | None:
    simple_names = (
        "Null",
        "Boolean",
        "UInt8",
        "UInt16",
        "UInt32",
        "UInt64",
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "Int128",
        "Float32",
        "Float64",
        "String",
        "Binary",
        "Date",
        "Time",
        "Object",
        "Unknown",
        "Categorical",
    )

    for name in simple_names:
        candidate = getattr(pl, name, None)
        if candidate is not None and dtype == candidate:
            return name
    return None


def _simple_dtype_from_name(name: str) -> pl.DataType | None:
    candidate = getattr(pl, name, None)
    if candidate is None:
        return None
    return candidate


def _dtype_to_python_payload(dtype: pl.DataType | type[pl.DataType]) -> object:
    if isinstance(dtype, type):
        return {"kind": "class", "name": dtype.__name__}

    simple_name = _simple_dtype_name(dtype)
    if simple_name is not None:
        return {"kind": "class", "name": simple_name}

    if isinstance(dtype, pl.List):
        return {
            "kind": "list",
            "inner": _dtype_to_python_payload(dtype.inner),
        }

    if isinstance(dtype, pl.Array):
        return {
            "kind": "array",
            "inner": _dtype_to_python_payload(dtype.inner),
            "shape": list(dtype.shape),
        }

    if isinstance(dtype, pl.Struct):
        return {
            "kind": "struct",
            "fields": [
                {
                    "name": field.name,
                    "dtype": _dtype_to_python_payload(field.dtype),
                }
                for field in dtype.fields
            ],
        }

    if isinstance(dtype, pl.Datetime):
        return {
            "kind": "datetime",
            "time_unit": dtype.time_unit,
            "time_zone": dtype.time_zone,
        }

    if isinstance(dtype, pl.Duration):
        return {
            "kind": "duration",
            "time_unit": dtype.time_unit,
        }

    if isinstance(dtype, pl.Decimal):
        return {
            "kind": "decimal",
            "precision": dtype.precision,
            "scale": dtype.scale,
        }

    if isinstance(dtype, pl.Enum):
        categories = getattr(dtype, "categories", None)
        return {
            "kind": "enum",
            "categories": list(categories) if categories is not None else None,
        }

    return {
        "kind": "string_repr",
        "value": str(dtype),
    }


def _dtype_from_python_payload(payload: object) -> pl.DataType:
    if not isinstance(payload, dict):
        raise TypeError(f"Expected dict fallback payload for DataType, got {type(payload)!r}")

    kind = payload.get("kind")
    if not isinstance(kind, str):
        raise TypeError("DataType payload missing string field 'kind'")

    if kind == "class":
        name = payload.get("name")
        if not isinstance(name, str):
            raise TypeError("DataType payload missing string field 'name'")
        dtype = _simple_dtype_from_name(name)
        if dtype is None:
            raise ValueError(f"Unsupported simple Polars dtype: {name!r}")
        return dtype

    if kind == "list":
        return pl.List(_dtype_from_python_payload(payload["inner"]))

    if kind == "array":
        inner = _dtype_from_python_payload(payload["inner"])
        shape = payload.get("shape")
        if isinstance(shape, list):
            shape_tuple = tuple(int(x) for x in shape)
            if len(shape_tuple) == 1:
                return pl.Array(inner, shape_tuple[0])
            return pl.Array(inner, shape_tuple)
        raise TypeError("Array dtype payload missing list field 'shape'")

    if kind == "struct":
        raw_fields = payload.get("fields")
        if not isinstance(raw_fields, list):
            raise TypeError("Struct dtype payload missing list field 'fields'")
        return pl.Struct(
            [
                pl.Field(
                    field["name"],
                    _dtype_from_python_payload(field["dtype"]),
                )
                for field in raw_fields
            ]
        )

    if kind == "datetime":
        return pl.Datetime(
            time_unit=payload.get("time_unit"),
            time_zone=payload.get("time_zone"),
        )

    if kind == "duration":
        return pl.Duration(
            time_unit=payload.get("time_unit"),
        )

    if kind == "decimal":
        return pl.Decimal(
            precision=payload.get("precision"),
            scale=payload.get("scale"),
        )

    if kind == "enum":
        categories = payload.get("categories")
        if categories is None:
            raise ValueError("Enum dtype payload missing categories")
        if not isinstance(categories, list):
            raise TypeError("Enum dtype payload field 'categories' must be a list")
        return pl.Enum(categories)

    if kind == "string_repr":
        value = payload.get("value")
        if not isinstance(value, str):
            raise TypeError("DataType payload missing string field 'value'")

        # minimal compatibility escape hatch
        simple = _simple_dtype_from_name(value)
        if simple is not None:
            return simple

        raise ValueError(f"Unsupported Polars dtype string representation: {value!r}")

    raise ValueError(f"Unsupported Polars dtype payload kind: {kind!r}")


def _coerce_dtype(obj: object) -> pl.DataType | type[pl.DataType] | None:
    if isinstance(obj, type):
        module = getattr(obj, "__module__", "")
        if module.startswith("polars"):
            return obj

    if isinstance(obj, str):
        return _simple_dtype_from_name(obj)

    try:
        payload = _dtype_to_python_payload(obj)  # type: ignore[arg-type]
        _ = _dtype_from_python_payload(payload)
        return obj  # type: ignore[return-value]
    except Exception:
        return None


def _dataframe_to_python_payload(df: pl.DataFrame) -> dict[str, object]:
    return {
        "columns": [
            {
                "name": name,
                "values": df.get_column(name).to_list(),
            }
            for name in df.columns
        ],
        "schema": [
            {
                "name": name,
                "dtype": _dtype_to_python_payload(dtype),
            }
            for name, dtype in df.schema.items()
        ],
    }


def _dataframe_from_python_payload(payload: object) -> pl.DataFrame:
    if not isinstance(payload, dict):
        raise TypeError(f"Expected dict fallback payload for DataFrame, got {type(payload)!r}")

    column_entries = payload.get("columns")
    schema_entries = payload.get("schema")

    if not isinstance(column_entries, list):
        raise TypeError("DataFrame fallback payload missing list 'columns'")

    data: dict[str, list[object]] = {}
    column_order: list[str] = []

    for entry in column_entries:
        if not isinstance(entry, dict):
            raise TypeError("DataFrame fallback payload column entry must be a dict")
        if "name" not in entry or "values" not in entry:
            raise TypeError(
                "DataFrame fallback payload column entry must contain 'name' and 'values'"
            )

        name = entry["name"]
        values = entry["values"]

        if not isinstance(name, str):
            raise TypeError("DataFrame fallback payload column 'name' must be a str")
        if not isinstance(values, list):
            raise TypeError("DataFrame fallback payload column 'values' must be a list")

        data[name] = values
        column_order.append(name)

    schema: dict[str, pl.DataType] | None = None
    if isinstance(schema_entries, list):
        maybe_schema: dict[str, pl.DataType] = {}
        try:
            for entry in schema_entries:
                if not isinstance(entry, dict):
                    raise TypeError("DataFrame fallback schema entry must be a dict")

                name = entry.get("name")
                dtype_payload = entry.get("dtype")

                if not isinstance(name, str):
                    raise TypeError("DataFrame fallback schema field 'name' must be a str")

                maybe_schema[name] = _dtype_from_python_payload(dtype_payload)
            schema = maybe_schema
        except Exception:
            schema = None

    if schema:
        try:
            return pl.DataFrame(data, schema=schema).select(column_order)
        except Exception:
            pass

    return pl.DataFrame(data).select(column_order)


def _series_to_python_payload(series: pl.Series) -> dict[str, object]:
    return {
        "name": series.name,
        "values": series.to_list(),
        "dtype": _dtype_to_python_payload(series.dtype),
    }


def _series_from_python_payload(payload: object) -> pl.Series:
    if not isinstance(payload, dict):
        raise TypeError(f"Expected dict fallback payload for Series, got {type(payload)!r}")

    name = payload.get("name")
    values = payload.get("values")
    dtype_payload = payload.get("dtype")

    if name is not None and not isinstance(name, str):
        raise TypeError("Series fallback payload field 'name' must be a str or None")
    if not isinstance(values, list):
        raise TypeError("Series fallback payload missing list 'values'")

    kwargs: dict[str, object] = {"name": name, "values": values}

    if dtype_payload is not None:
        try:
            kwargs["dtype"] = _dtype_from_python_payload(dtype_payload)
        except Exception:
            pass

    try:
        return pl.Series(**kwargs)
    except Exception:
        return pl.Series(name=name, values=values)


def _lazyframe_to_bytes(lf: pl.LazyFrame) -> bytes:
    data = lf.serialize(format="binary")
    if isinstance(data, str):
        return data.encode("utf-8")
    return bytes(data)


def _lazyframe_from_bytes(data: bytes) -> pl.LazyFrame:
    return pl.LazyFrame.deserialize(StdBytesIO(data), format="binary")


def _expr_to_bytes(expr: pl.Expr) -> bytes:
    data = expr.meta.serialize(format="binary")
    if isinstance(data, str):
        return data.encode("utf-8")
    return bytes(data)


def _expr_from_bytes(data: bytes) -> pl.Expr:
    return pl.Expr.deserialize(StdBytesIO(data), format="binary")


def _schema_items_from_object(obj: object) -> list[tuple[str, pl.DataType | type[pl.DataType]]] | None:
    if isinstance(obj, pl.Schema):
        return list(obj.items())

    if isinstance(obj, Mapping):
        items: list[tuple[str, pl.DataType | type[pl.DataType]]] = []
        for key, value in obj.items():
            if not isinstance(key, str):
                return None
            dtype = _coerce_dtype(value)
            if dtype is None:
                return None
            items.append((key, dtype))
        return items

    return None


def _schema_to_python_payload(schema: object) -> dict[str, object]:
    items = _schema_items_from_object(schema)
    if items is None:
        raise TypeError(f"Unsupported schema object: {type(schema)!r}")

    return {
        "items": [
            {
                "name": name,
                "dtype": _dtype_to_python_payload(dtype),
            }
            for name, dtype in items
        ]
    }


def _schema_from_python_payload(payload: object) -> pl.Schema:
    if not isinstance(payload, dict):
        raise TypeError(f"Expected dict fallback payload for Schema, got {type(payload)!r}")

    raw_items = payload.get("items")
    if not isinstance(raw_items, list):
        raise TypeError("Schema fallback payload missing list 'items'")

    items: list[tuple[str, pl.DataType]] = []
    for entry in raw_items:
        if not isinstance(entry, dict):
            raise TypeError("Schema fallback payload entry must be a dict")

        name = entry.get("name")
        dtype_payload = entry.get("dtype")

        if not isinstance(name, str):
            raise TypeError("Schema fallback payload field 'name' must be a str")

        dtype = _dtype_from_python_payload(dtype_payload)
        items.append((name, dtype))

    return pl.Schema(items)


@dataclass(frozen=True, slots=True)
class PolarsSerialized(Serialized[TPolars], Generic[TPolars]):
    TAG: ClassVar[int]

    @property
    def value(self) -> TPolars:
        raise NotImplementedError

    def as_python(self) -> TPolars:
        return self.value

    def decode_arrow_buffer(self) -> pa.Buffer:
        return pa.py_buffer(self.decode())

    @staticmethod
    def _deserialize_nested_payload(data: bytes) -> object:
        nested = Serialized.read_from(BytesIO(data), pos=0)
        return nested.as_python()

    @staticmethod
    def _serialize_nested_payload(
        payload: object,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        return Serialized.from_python_object(
            payload,
            metadata=metadata,
            codec=codec,
        )

    @staticmethod
    def _nested_serialized_bytes(nested: Serialized[object]) -> bytes:
        buf = BytesIO()
        nested.write_to(buf)
        return buf.to_bytes()

    @classmethod
    def from_python_object(
        cls,
        obj: object,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object] | None:
        if isinstance(obj, pl.DataFrame):
            return PolarsDataFrameSerialized.from_value(
                obj,
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, pl.Series):
            return PolarsSeriesSerialized.from_value(
                obj,
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, pl.LazyFrame):
            return PolarsLazyFrameSerialized.from_value(
                obj,
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, pl.Expr):
            return PolarsExprSerialized.from_value(
                obj,
                metadata=metadata,
                codec=codec,
            )

        if isinstance(obj, pl.Schema) or isinstance(obj, Mapping):
            items = _schema_items_from_object(obj)
            if items is not None:
                return PolarsSchemaSerialized.from_value(
                    obj,
                    metadata=metadata,
                    codec=codec,
                )

        dtype = _coerce_dtype(obj)
        if dtype is not None:
            return PolarsDataTypeSerialized.from_value(
                dtype,
                metadata=metadata,
                codec=codec,
            )

        return None


@dataclass(frozen=True, slots=True)
class PolarsDataFrameSerialized(PolarsSerialized[pl.DataFrame]):
    TAG: ClassVar[int] = Tags.POLARS_DATAFRAME

    @property
    def value(self) -> pl.DataFrame:
        metadata = self.metadata or {}
        serialization_format = metadata.get(b"serialization_format", b"arrow")

        if serialization_format == b"python_serialized":
            payload = self._deserialize_nested_payload(self.decode())
            return _dataframe_from_python_payload(payload)

        table = _table_from_ipc_file_buffer(self.decode_arrow_buffer())
        return _arrow_table_to_dataframe(table)

    @classmethod
    def from_value(
        cls,
        df: pl.DataFrame,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        try:
            table = _dataframe_to_arrow_table(df)
            buf = _table_to_ipc_file_buffer(table, metadata=None)
            return cls.build(
                tag=cls.TAG,
                data=buf,
                metadata=_merge_metadata(metadata, {b"serialization_format": b"arrow"}),
                codec=codec,
            )
        except Exception:
            payload = _dataframe_to_python_payload(df)
            nested = cls._serialize_nested_payload(
                payload,
                metadata=None,
                codec=codec,
            )
            return cls.build(
                tag=cls.TAG,
                data=cls._nested_serialized_bytes(nested),
                metadata=_merge_metadata(
                    metadata,
                    {b"serialization_format": b"python_serialized"},
                ),
                codec=codec,
            )


@dataclass(frozen=True, slots=True)
class PolarsSeriesSerialized(PolarsSerialized[pl.Series]):
    TAG: ClassVar[int] = Tags.POLARS_SERIES

    @property
    def value(self) -> pl.Series:
        metadata = self.metadata or {}
        serialization_format = metadata.get(b"serialization_format", b"arrow")

        if serialization_format == b"python_serialized":
            payload = self._deserialize_nested_payload(self.decode())
            return _series_from_python_payload(payload)

        table = _table_from_ipc_file_buffer(self.decode_arrow_buffer())
        return _arrow_table_to_series(table)

    @classmethod
    def from_value(
        cls,
        series: pl.Series,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        try:
            table = _series_to_arrow_table(series)
            buf = _table_to_ipc_file_buffer(table, metadata=None)
            return cls.build(
                tag=cls.TAG,
                data=buf,
                metadata=_merge_metadata(metadata, {b"serialization_format": b"arrow"}),
                codec=codec,
            )
        except Exception:
            payload = _series_to_python_payload(series)
            nested = cls._serialize_nested_payload(
                payload,
                metadata=None,
                codec=codec,
            )
            return cls.build(
                tag=cls.TAG,
                data=cls._nested_serialized_bytes(nested),
                metadata=_merge_metadata(
                    metadata,
                    {b"serialization_format": b"python_serialized"},
                ),
                codec=codec,
            )


@dataclass(frozen=True, slots=True)
class PolarsLazyFrameSerialized(PolarsSerialized[pl.LazyFrame]):
    TAG: ClassVar[int] = Tags.POLARS_LAZYFRAME

    @property
    def value(self) -> pl.LazyFrame:
        return _lazyframe_from_bytes(self.decode())

    @classmethod
    def from_value(
        cls,
        lf: pl.LazyFrame,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        payload = _lazyframe_to_bytes(lf)
        return cls.build(
            tag=cls.TAG,
            data=payload,
            metadata=_merge_metadata(
                metadata,
                {b"serialization_format": b"polars_binary"},
            ),
            codec=codec,
        )


@dataclass(frozen=True, slots=True)
class PolarsExprSerialized(PolarsSerialized[pl.Expr]):
    TAG: ClassVar[int] = Tags.POLARS_EXPR

    @property
    def value(self) -> pl.Expr:
        return _expr_from_bytes(self.decode())

    @classmethod
    def from_value(
        cls,
        expr: pl.Expr,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        payload = _expr_to_bytes(expr)
        return cls.build(
            tag=cls.TAG,
            data=payload,
            metadata=_merge_metadata(
                metadata,
                {b"serialization_format": b"polars_binary"},
            ),
            codec=codec,
        )


@dataclass(frozen=True, slots=True)
class PolarsSchemaSerialized(PolarsSerialized[pl.Schema]):
    TAG: ClassVar[int] = Tags.POLARS_SCHEMA

    @property
    def value(self) -> pl.Schema:
        payload = self._deserialize_nested_payload(self.decode())
        return _schema_from_python_payload(payload)

    @classmethod
    def from_value(
        cls,
        schema: object,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        payload = _schema_to_python_payload(schema)
        nested = cls._serialize_nested_payload(
            payload,
            metadata=None,
            codec=codec,
        )
        return cls.build(
            tag=cls.TAG,
            data=cls._nested_serialized_bytes(nested),
            metadata=_merge_metadata(
                metadata,
                {b"serialization_format": b"python_serialized"},
            ),
            codec=codec,
        )


@dataclass(frozen=True, slots=True)
class PolarsDataTypeSerialized(
    PolarsSerialized[pl.DataType | type[pl.DataType]]
):
    TAG: ClassVar[int] = Tags.POLARS_DATATYPE

    @property
    def value(self) -> pl.DataType:
        payload = self._deserialize_nested_payload(self.decode())
        return _dtype_from_python_payload(payload)

    @classmethod
    def from_value(
        cls,
        dtype: object,
        *,
        metadata: Mapping[bytes, bytes] | None = None,
        codec: int | None = None,
    ) -> Serialized[object]:
        coerced = _coerce_dtype(dtype)
        if coerced is None:
            raise TypeError(f"Unsupported Polars dtype object: {type(dtype)!r}")

        nested = cls._serialize_nested_payload(
            _dtype_to_python_payload(coerced),
            metadata=None,
            codec=codec,
        )
        return cls.build(
            tag=cls.TAG,
            data=cls._nested_serialized_bytes(nested),
            metadata=_merge_metadata(
                metadata,
                {b"serialization_format": b"python_serialized"},
            ),
            codec=codec,
        )


for cls in PolarsSerialized.__subclasses__():
    Tags.register_class(cls, tag=cls.TAG)



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from typing import Mapping, Optional

from urllib3 import BaseHTTPResponse

from ..buffer import BytesIO
from ..request import PreparedRequest
from ..response import Response

__all__ = [
    "HTTPResponse"
]


@dataclass(slots=True)
class HTTPResponse(Response):

    @classmethod
    def from_urllib3(
        cls,
        request: "PreparedRequest",
        response: BaseHTTPResponse,
        tags: Optional[Mapping[str, str]],
        received_at: dt.datetime
    ) -> "HTTPResponse":
        response = cls(
            request=request,
            status_code=response.status,
            headers=dict(response.headers),
            buffer=BytesIO(),
            tags=tags,
            received_at=received_at
        )

        if request.prepare_response is not None:
            return request.prepare_response(response)

        return response

    def drain_urllib3(
        self,
        response: BaseHTTPResponse,
        stream: bool,
        *,
        amt: int = 512 * 1024,
        release_conn: bool = True
    ) -> BytesIO:
        buffer = self.buffer

        try:
            if stream:
                for chunk in response.stream(amt=amt):
                    buffer.write(chunk)
            else:
                buffer.write(response.read())
        finally:
            if release_conn:
                response.release_conn()

        return self


"""Spill-to-disk byte buffer with transparent memory/path backing.

Design
------
- Own logical cursor (self._pos), independent from any OS/file cursor
- In-memory backing is bytearray + logical size
- Spilled backing is a filesystem path only (no persistent Python file handle)
- Uses os.pread/os.pwrite for cursorless file IO
- Keeps optional readonly mmap for zero-copy reads when spilled

Semantics
---------
- By default, construction is deep-copying for raw bytes/path/file-like inputs
- ``BytesIO(BytesIO(...), copy=True)`` duplicates content
- ``BytesIO(BytesIO(...), copy=False)`` aliases spilled-path backing only
- ``BytesIO(path, copy=True)`` copies file content into owned backing
- ``BytesIO(path, copy=False)`` aliases the original path and mutates it
- ``BytesIO(io.BytesIO(...))`` and other file-like sources read from the current cursor onward
"""

from __future__ import annotations

import base64
import io
import mmap
import os
import shutil
import struct
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, IO, Literal, Optional, Union

import yggdrasil.pickle.json as json_module
from yggdrasil.io.config import DEFAULT_CONFIG, BufferConfig
from yggdrasil.io.enums import MediaType, MimeType
from yggdrasil.io.types import BytesLike

if TYPE_CHECKING:
    import blake3
    import xxhash

    from .media_io import MediaIO
    from ..enums import Codec

__all__ = ["BytesIO", "BufferLike"]

BufferLike = Union[
    bytes,
    bytearray,
    memoryview,
    io.BytesIO,
    "BytesIO",
    str,
    Path,
    IO[bytes],
]

_HEAD_DEFAULT = 128
_COPY_CHUNK_SIZE = 8 * 1024 * 1024
_HAS_PREAD = hasattr(os, "pread")
_HAS_PWRITE = hasattr(os, "pwrite")


class BytesIO(io.RawIOBase):
    __slots__ = (
        "_cfg",
        "_buf",
        "_size",
        "_pos",
        "_path",
        "_mmap",
        "_closed",
        "_owns_path",
        "_media_type",
    )

    def __init__(
        self,
        data: IO[bytes] | bytes | bytearray | memoryview | str | Path | None = None,
        *,
        media_type: Optional["MediaType"] = None,
        config: BufferConfig | None = None,
        copy: bool = False,
    ) -> None:
        super().__init__()
        self._cfg: BufferConfig = config or DEFAULT_CONFIG

        self._buf: bytearray | None = None
        self._size: int = 0
        self._pos: int = 0

        self._path: Path | None = None
        self._mmap: mmap.mmap | None = None
        self._owns_path: bool = False
        self._media_type: MediaType | None = media_type

        self._closed: bool = False
        self._init_from(data, copy=copy)

    def __bool__(self) -> bool:
        return self.size > 0

    def __bytes__(self) -> bytes:
        return self.to_bytes()

    def __iter__(self):
        return iter(self.to_bytes())

    def __len__(self) -> int:
        return self.size

    def __repr__(self) -> str:
        if self._closed:
            return "<BytesIO [closed]>"
        state = "spilled" if self.spilled else "memory"
        return f"<BytesIO [{state}] size={self.size} bytes pos={self._pos}>"

    def __enter__(self) -> "BytesIO":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def __del__(self) -> None:
        try:
            self.close()
        except Exception:
            pass

    def __getstate__(self):
        if self.size > 1024 * 1024:
            from yggdrasil.io import ZSTD

            blob = self.compress(codec=ZSTD, copy=True).to_bytes()
            codec = ZSTD.name
        else:
            blob = self.to_bytes()
            codec = None

        return {
            "data": blob,
            "codec": codec,
            "media_type": self._media_type,
        }

    def __setstate__(self, state):
        blob = state["data"]
        codec = state.get("codec")
        media_type = state.get("media_type")

        if codec is not None:
            from yggdrasil.io import Codec
            codec = Codec.parse(codec)
            blob = codec.decompress_bytes(blob)

        self.__init__(blob, copy=False, media_type=media_type)

    # ------------------------------------------------------------------
    # Init
    # ------------------------------------------------------------------

    def _init_from(self, data: Any, *, copy: bool) -> None:
        if data is None:
            self._buf = bytearray()
            self._size = 0
            self._pos = 0
            self._path = None
            self._owns_path = False
            return

        if isinstance(data, (bytes, bytearray, memoryview)):
            self._init_from_bytes(memoryview(data))
            return

        if isinstance(data, io.BytesIO):
            self._init_from_stdlib_bytesio(data)
            return

        if isinstance(data, (str, Path)):
            self._init_from_path(Path(data), copy=copy)
            return

        if isinstance(data, BytesIO):
            self._init_from_bytesio(data, copy=copy)
            return

        if hasattr(data, "read"):
            self._init_from_filelike(data)
            return

        raise TypeError(
            f"{type(self).__name__} does not accept data of type {type(data)!r}. "
            "Pass bytes/bytearray/memoryview, io.BytesIO, BytesIO, a file-like object, or a Path."
        )

    def _init_from_bytes(self, mv: memoryview) -> None:
        n = len(mv)
        if n > self._cfg.spill_bytes:
            self._spill_from_bytes(mv)
        else:
            self._buf = bytearray(mv.tobytes() if not mv.contiguous else mv)
            self._size = n
            self._path = None
            self._owns_path = False
        self._pos = 0

    def _init_from_stdlib_bytesio(self, src: io.BytesIO) -> None:
        self.pos = src.tell()

        self._buf = bytearray(src.getvalue())
        self._size = len(self._buf)
        self._path = None

        if self._size > self._cfg.spill_bytes:
            self.spill_to_file()

    def _init_from_bytesio(self, src: BytesIO, *, copy: bool) -> None:
        if copy:
            self._init_from_bytes(memoryview(src.to_bytes()))
            self._pos = 0
            return

        # Aliasing memory-backed instances is unsafe because metadata
        # (_size/_pos) would diverge between aliases. Copy instead.
        if src._buf is not None:
            self._init_from_bytes(memoryview(src.to_bytes()))
            self._pos = src._pos
            return

        self._buf = None
        self._size = 0
        self._path = src._path
        self._owns_path = False
        self._pos = src._pos
        self._media_type = src._media_type

    def _init_from_path(self, path: Path, *, copy: bool) -> None:
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()

        if copy:
            size = path.stat().st_size
            if size > self._cfg.spill_bytes:
                dst = self._cfg.create_spill_path()
                shutil.copyfile(path, dst)
                self._buf = None
                self._size = 0
                self._path = dst
                self._owns_path = True
            else:
                payload = path.read_bytes()
                self._buf = bytearray(payload)
                self._size = len(payload)
                self._path = None
                self._owns_path = False
        else:
            self._buf = None
            self._size = 0
            self._path = path
            self._owns_path = False

        self._pos = 0

    def _init_from_filelike(self, src: Any) -> None:
        if hasattr(src, "seek") and hasattr(src, "tell"):
            self.pos = src.tell()
            src.seek(0, io.SEEK_END)
            end = src.tell()
            src.seek(self.pos)
            remaining = max(0, end - self.pos)

            if remaining > self._cfg.spill_bytes:
                path = self._cfg.create_spill_path()
                with path.open("w+b") as fh:
                    shutil.copyfileobj(src, fh, length=_COPY_CHUNK_SIZE)
                    fh.flush()
                self._buf = None
                self._size = 0
                self._path = path
                self._owns_path = True
            else:
                payload = src.read()
                self._buf = bytearray(payload)
                self._size = len(payload)
                self._path = None
                self._owns_path = False
        else:
            payload = src.read()
            self._init_from_bytes(memoryview(payload))
            self._pos = 0
            return

    # ------------------------------------------------------------------
    # Backing helpers
    # ------------------------------------------------------------------

    def _create_spill_path(self) -> Path:
        if self._path is None:
            self._path = self._cfg.create_spill_path()
        return self._path

    def _invalidate_mmap(self) -> None:
        if self._mmap is not None:
            try:
                self._mmap.close()
            except Exception:
                pass
            self._mmap = None

    def _spill_from_bytes(self, mv: memoryview) -> None:
        path = self._cfg.create_spill_path()
        with path.open("w+b") as fh:
            fh.write(mv.tobytes() if not mv.contiguous else mv)
            fh.flush()

        self._buf = None
        self._size = 0
        self._path = path
        self._owns_path = True
        self._invalidate_mmap()

    def spill_to_file(self) -> None:
        if self._buf is None:
            return

        cfg = self._cfg
        tmp_dir = str(cfg.tmp_dir) if cfg.tmp_dir is not None else None
        fd, name = tempfile.mkstemp(prefix=cfg.prefix, suffix=cfg.suffix, dir=tmp_dir)
        path = Path(name)

        try:
            if self._size:
                os.write(fd, memoryview(self._buf)[: self._size])
        finally:
            os.close(fd)

        self._buf = None
        self._size = 0
        self._path = path
        self._owns_path = True
        self._invalidate_mmap()

    def _reset_backing_keep_open(self) -> None:
        self._invalidate_mmap()

        old_path = self._path
        old_owns_path = self._owns_path

        self._path = None
        self._owns_path = False

        if old_path is not None and old_owns_path and not self._cfg.keep_spilled_file:
            try:
                old_path.unlink(missing_ok=True)
            except Exception:
                pass

        self._buf = bytearray()
        self._size = 0
        self._pos = 0
        self._media_type = None

    def _replace_with_payload(self, payload: bytes) -> None:
        self._reset_backing_keep_open()
        self._init_from(payload, copy=False)

    def buffer(self) -> IO[bytes]:
        """
        Return a file object for spilled backing.

        This is primarily for internal helpers that need truncate/flush/fd access.
        """
        if self._buf is not None:
            return io.BytesIO(bytes(memoryview(self._buf)[: self._size]))
        if self._path is None:
            raise RuntimeError("No backing store available")
        return self._path.open("r+b")

    # ------------------------------------------------------------------
    # Cursorless IO primitives
    # ------------------------------------------------------------------

    def pread(self, n: int, pos: int) -> bytes:
        if n <= 0:
            return b""

        if pos < 0:
            raise ValueError("pread position must be >= 0")

        if self._buf is not None:
            end = min(pos + n, self._size)
            if pos >= end:
                return b""
            return bytes(memoryview(self._buf)[pos:end])

        if self._path is None:
            return b""

        with self._path.open("rb") as fh:
            if _HAS_PREAD:
                return os.pread(fh.fileno(), n, pos)

            fh.seek(pos)
            return fh.read(n)

    def pwrite(self, mv: memoryview, pos: int) -> int:
        if len(mv) == 0:
            return 0
        if pos < 0:
            raise ValueError("pwrite position must be >= 0")

        if self._buf is not None:
            need = pos + len(mv)
            if need > len(self._buf):
                new_cap = max(need, int(len(self._buf) * 1.5) + 1)
                self._buf.extend(b"\x00" * (new_cap - len(self._buf)))

            memoryview(self._buf)[pos : pos + len(mv)] = mv
            self._size = max(self._size, need)
            return len(mv)

        if self._path is None:
            raise RuntimeError("No backing store available for write")

        with self._path.open("r+b") as fh:
            if _HAS_PWRITE:
                written = int(os.pwrite(fh.fileno(), mv, pos))
            else:
                fh.seek(pos)
                out = fh.write(mv)
                fh.flush()
                written = len(mv) if out is None else int(out)

        return written

    def _ensure_spill_for_growth(self, extra: int) -> None:
        if self._buf is None:
            return
        projected = max(self._size, self._pos) + extra
        if projected > self._cfg.spill_bytes:
            self.spill_to_file()

    # ------------------------------------------------------------------
    # Factory
    # ------------------------------------------------------------------

    @classmethod
    def parse(cls, obj: Any, config: BufferConfig | None = None) -> "BytesIO":
        if isinstance(obj, cls):
            return obj
        return cls(obj, config=config)

    # ------------------------------------------------------------------
    # Introspection
    # ------------------------------------------------------------------

    @property
    def config(self) -> BufferConfig:
        return self._cfg

    @property
    def spilled(self) -> bool:
        return self._path is not None

    @property
    def path(self) -> Path | None:
        return self._path

    @property
    def size(self) -> int:
        if self._buf is not None:
            return self._size
        if self._path is not None and self._path.exists():
            return self._path.stat().st_size
        return 0

    @property
    def media_type(self) -> MediaType:
        if self._media_type is None:
            self._media_type = MediaType.parse(self, default=MediaType(MimeType.OCTET_STREAM))
        return self._media_type

    @media_type.setter
    def media_type(self, value: "MediaType"):
        self.set_media_type(value, safe=True)

    def set_media_type(
        self,
        value: MediaType,
        *,
        safe: bool = True,
    ) -> "BytesIO":
        from ..enums import MediaType as _MediaType

        parsed = _MediaType.parse(value)
        if parsed is None and safe:
            raise ValueError(f"Invalid media type: {value!r}")
        self._media_type = parsed
        return self

    def media_io(
        self,
        media: Optional[MediaType] = None,
    ) -> "MediaIO":
        from .media_io import MediaIO

        media = self.media_type if media is None else MediaType.parse(media)
        return MediaIO.make(buffer=self, media=media)

    # ------------------------------------------------------------------
    # RawIOBase-ish API
    # ------------------------------------------------------------------

    def json_load(
        self,
        orient: Optional[Literal["records", "split", "index", "columns", "values"]] = None,
        *,
        media_type: Optional[MediaType] = None,
    ):
        media_type = media_type or self.media_type

        if media_type.codec is None:
            if media_type.is_json:
                with self.view(pos=0) as v:
                    return json_module.load(v)

            mio = self.media_io(media_type)

            if orient == "columns":
                return mio.read_pydict()
            return mio.read_pylist()

        with self.decompress(codec=media_type.codec, copy=True) as decompressed:
            return decompressed.json_load(
                orient=orient,
                media_type=media_type.with_codec(None),
            )

    def exists(self) -> bool:
        return self.size > 0

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def flush(self) -> None:
        return None

    def head(self, n: int = _HEAD_DEFAULT):
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")
        if n <= 0 or self.size == 0:
            return memoryview(b"")

        mv = self.memoryview()
        return mv[: min(n, len(mv))]

    def tell(self) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")
        return int(self._pos)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if whence == io.SEEK_SET:
            new_pos = int(offset)
        elif whence == io.SEEK_CUR:
            new_pos = self._pos + int(offset)
        elif whence == io.SEEK_END:
            new_pos = self.size + int(offset)
        else:
            raise ValueError(f"Invalid whence: {whence!r}")

        if new_pos < 0:
            raise ValueError("Negative seek position")

        self._pos = new_pos
        return self._pos

    def read(self, size: int = -1) -> bytes:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if size is None or size < 0:
            size = max(0, self.size - self._pos)

        out = self.pread(size, self._pos)
        self._pos += len(out)
        return out

    def readinto(self, b) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        mv = memoryview(b)
        if len(mv) == 0:
            return 0

        chunk = self.pread(len(mv), self._pos)
        mv[: len(chunk)] = chunk
        self._pos += len(chunk)
        return len(chunk)

    def readinto1(self, b) -> int:
        return self.readinto(b)

    def write(self, b: Any, *, batch_size: int = 1024 * 1024) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if b is None:
            return 0

        if isinstance(b, str):
            return self.write_str(b)

        if isinstance(b, (bytes, bytearray, memoryview)):
            return self.write_bytes(b)

        if isinstance(b, (io.RawIOBase, io.BufferedIOBase)) or hasattr(b, "read"):
            total = 0
            while True:
                chunk = b.read(batch_size)
                if not chunk:
                    break
                total += self.write_bytes(chunk)
            return total

        return self.write_bytes(bytes(b))

    def write_linebreak(self, newline: str = "\n") -> int:
        return self.write(newline)

    def write_bytes(self, b: bytes | bytearray | memoryview) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        mv = memoryview(b)
        if len(mv) == 0:
            return 0

        self._ensure_spill_for_growth(len(mv))
        n = self.pwrite(mv, self._pos)
        self._pos += n

        if self._buf is None:
            self._invalidate_mmap()

        return n

    def write_str(self, s: str, encoding: str = "utf-8") -> int:
        if not s:
            return 0
        return self.write_bytes(s.encode(encoding))

    def truncate(self, size: int | None = None) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if size is None:
            size = self._pos
        size = int(size)

        if size < 0:
            raise ValueError("Negative size value")

        if self._buf is not None:
            if size < self._size:
                self._size = size
            else:
                if size > len(self._buf):
                    self._buf.extend(b"\x00" * (size - len(self._buf)))
                self._size = size

            if self._pos > size:
                self._pos = size
            return size

        if self._path is None:
            raise RuntimeError("No backing store available for truncate")

        with self._path.open("r+b") as fh:
            fh.truncate(size)

        if self._pos > size:
            self._pos = size
        self._invalidate_mmap()
        return size

    # ------------------------------------------------------------------
    # Structured binary I/O — little-endian
    # ------------------------------------------------------------------

    def _read_exact(self, n: int) -> bytes:
        data = self.read(n)
        if len(data) != n:
            raise EOFError(f"expected {n} bytes, got {len(data)}")
        return data

    def read_int8(self) -> int:
        return struct.unpack("<b", self._read_exact(1))[0]

    def write_int8(self, value: int) -> int:
        return self.write(struct.pack("<b", int(value)))

    def read_uint8(self) -> int:
        return struct.unpack("<B", self._read_exact(1))[0]

    def write_uint8(self, value: int) -> int:
        return self.write(struct.pack("<B", int(value)))

    def read_int16(self) -> int:
        return struct.unpack("<h", self._read_exact(2))[0]

    def write_int16(self, value: int) -> int:
        return self.write(struct.pack("<h", int(value)))

    def read_uint16(self) -> int:
        return struct.unpack("<H", self._read_exact(2))[0]

    def write_uint16(self, value: int) -> int:
        return self.write(struct.pack("<H", int(value)))

    def read_int32(self) -> int:
        return struct.unpack("<i", self._read_exact(4))[0]

    def write_int32(self, value: int) -> int:
        return self.write(struct.pack("<i", int(value)))

    def read_uint32(self) -> int:
        return struct.unpack("<I", self._read_exact(4))[0]

    def write_uint32(self, value: int) -> int:
        return self.write(struct.pack("<I", int(value)))

    def read_int64(self) -> int:
        return struct.unpack("<q", self._read_exact(8))[0]

    def write_int64(self, value: int) -> int:
        return self.write(struct.pack("<q", int(value)))

    def read_uint64(self) -> int:
        return struct.unpack("<Q", self._read_exact(8))[0]

    def write_uint64(self, value: int) -> int:
        return self.write(struct.pack("<Q", int(value)))

    def read_f32(self) -> float:
        return struct.unpack("<f", self._read_exact(4))[0]

    def write_f32(self, value: float) -> int:
        return self.write(struct.pack("<f", float(value)))

    def read_f64(self) -> float:
        return struct.unpack("<d", self._read_exact(8))[0]

    def write_f64(self, value: float) -> int:
        return self.write(struct.pack("<d", float(value)))

    def read_bool(self) -> bool:
        return bool(self.read_uint8())

    def write_bool(self, value: bool) -> int:
        return self.write_uint8(1 if value else 0)

    def read_bytes_u32(self) -> bytes:
        return self._read_exact(self.read_uint32())

    def write_bytes_u32(self, data: BytesLike) -> int:
        mv = memoryview(data)
        return self.write_uint32(len(mv)) + self.write_bytes(mv)

    def read_str_u32(self, encoding: str = "utf-8") -> str:
        return self.read_bytes_u32().decode(encoding)

    def write_str_u32(self, s: str, encoding: str = "utf-8") -> int:
        return self.write_bytes_u32(s.encode(encoding))

    # ------------------------------------------------------------------
    # Hashing
    # ------------------------------------------------------------------

    def xxh3_64(self) -> "xxhash.xxh3_64":
        from yggdrasil.xxhash import xxh3_64

        h = xxh3_64()
        h.update(self.memoryview())
        return h

    def xxh3_int64(self) -> int:
        u = self.xxh3_64().intdigest()
        return u if u < 2**63 else u - 2**64

    def blake3(self) -> "blake3.blake3":
        from yggdrasil.blake3 import blake3

        h = blake3(max_threads=blake3.AUTO)

        if self._buf is not None:
            if self._size:
                h.update(memoryview(self._buf)[: self._size])
            return h

        if self._path is not None:
            h.update_mmap(str(self._path))
            return h

        return h

    # ------------------------------------------------------------------
    # Convenience
    # ------------------------------------------------------------------

    def decode(self, encoding: str = "utf-8", errors: str = "replace") -> str:
        if self.size == 0:
            return ""
        return self.to_bytes().decode(encoding, errors)

    def getvalue(self) -> bytes:
        return self.to_bytes()

    def view(
        self,
        *,
        pos: Optional[int] = None,
        size: int | None = None,
        max_size: int | None = None,
    ):
        from .bytes_view import BytesIOView

        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if pos is None:
            pos = self._pos

            if pos == self.size:
                pos = 0

        pos = int(pos)
        if pos < 0:
            raise ValueError("view pos must be >= 0")

        if size is None:
            size = max(0, self.size - pos)
        else:
            size = int(size)
            if size < 0:
                raise ValueError("view length must be >= 0")

        return BytesIOView(
            parent=self,
            start=pos,
            size=size,
            pos=0,
            max_size=max_size,
        )

    def memoryview(self):
        if self._buf is not None:
            return memoryview(self._buf)[: self._size]

        if self._path is None or not self._path.exists():
            return memoryview(b"")

        size = self._path.stat().st_size
        if size == 0:
            return memoryview(b"")

        if self._mmap is None or self._mmap.closed:
            with self._path.open("rb") as fh:
                self._mmap = mmap.mmap(fh.fileno(), length=0, access=mmap.ACCESS_READ)

        return memoryview(self._mmap)

    def to_bytes(self) -> bytes:
        if self._buf is not None:
            return bytes(memoryview(self._buf)[: self._size])
        if self.size == 0:
            return b""
        return self.pread(self.size, 0)

    def to_base64(
        self,
        urlsafe: bool = True
    ) -> str:
        b = self.to_bytes()

        if urlsafe:
            return base64.urlsafe_b64encode(b).decode("ascii")
        else:
            return base64.b64encode(b).decode("ascii")

    def open_reader(self) -> IO[bytes]:
        if self._buf is not None:
            return io.BytesIO(bytes(memoryview(self._buf)[: self._size]))
        if self._path is None:
            raise RuntimeError("Spilled buffer has no path")
        return self._path.open("rb")

    def to_arrow_io(self, mode: str = "r"):
        from yggdrasil.arrow.lib import pyarrow as pa

        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if "r" in mode:
            if self.spilled and self._path is not None:
                return pa.memory_map(str(self._path), "r")

            mv = self.memoryview()
            buf = pa.py_buffer(mv) if len(mv) else pa.py_buffer(b"")
            return pa.BufferReader(buf)

        if "w" in mode or "a" in mode:
            if not self.spilled:
                self.spill_to_file()

            if self._path is None:
                raise RuntimeError("Failed to materialize spill file for Arrow IO")

            if "w" in mode:
                return pa.OSFile(str(self._path), mode="w")
            return pa.OSFile(str(self._path), mode="a")

        raise ValueError(f"Unsupported mode for to_arrow_io: {mode!r}")

    # ------------------------------------------------------------------
    # Compression helpers
    # ------------------------------------------------------------------

    def compress(
        self,
        codec: "Codec | str",
        *,
        copy: bool = False
    ) -> "BytesIO":
        from ..enums.codec import Codec as _Codec

        c = _Codec.parse(codec)
        if c is None:
            raise ValueError(f"Unknown codec: {codec!r}")

        mt = self.media_type.with_codec(c) if self._media_type else None
        payload = c.compress(self)
        payload._media_type = mt

        if copy:
            return payload

        self._replace_with_payload(payload)
        return self

    def decompress(
        self,
        codec: "Codec | str | None" = "infer",
        *,
        copy: bool = False
    ) -> "BytesIO":
        from ..enums.codec import Codec as _Codec

        if self._closed:
            raise ValueError("I/O operation on closed BytesIO")

        if codec == "infer":
            mt = self.media_type.without_codec()
            codec = self.media_type.codec
        else:
            mt = self._media_type.without_codec() if self._media_type else None
            codec = _Codec.parse(codec)

        if codec is None:
            return self

        raw = codec.decompress_bytes(self.to_bytes())

        if copy:
            payload = self.__class__(raw, copy=False, media_type=mt)
            return payload

        self._replace_with_payload(raw)
        self._media_type = mt
        return self

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    def cleanup(self) -> None:
        self.close()

    def close(self) -> None:
        if self._closed:
            return

        self._invalidate_mmap()

        old_path = self._path
        old_owns_path = self._owns_path

        self._buf = None
        self._size = 0
        self._pos = 0
        self._path = None
        self._owns_path = False
        self._media_type = None
        self._closed = True

        if old_path is not None and old_owns_path and not self._cfg.keep_spilled_file:
            try:
                old_path.unlink(missing_ok=True)
            except Exception:
                pass

        super().close()